# -*- coding: utf-8 -*-
__all__ = ('ROSLaunchController',)

from types import TracebackType
from typing import Iterator, Optional, Type

import attr
import dockerblade


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ROSLaunchController:
    """Provides an interface to a roslaunch process.

    Attributes
    ----------
    filename: str
        The absolute path of the XML launch file used by this process.
    command: str
        The command string that was used by this process.
    popen: dockerblade.popen.Popen
        An interface to the underlying exec process for this roslaunch process.
    pid: Optional[int]
        The PID of the launch process inside the container, if known.
    """
    filename: str
    popen: dockerblade.popen.Popen = attr.ib(repr=False)

    def __enter__(self) -> 'ROSLaunchController':
        return self

    def __exit__(self,
                 ex_type: Optional[Type[BaseException]],
                 ex_val: Optional[BaseException],
                 ex_tb: Optional[TracebackType]
                 ) -> None:
        self.close()

    @property
    def stream(self) -> Iterator[str]:
        yield from self.popen.stream()  # type: ignore

    @property
    def running(self) -> bool:
        """Checks whether or not this roslaunch process is still running."""
        return not self.popen.finished

    def terminate(self) -> None:
        """Terminates this roslaunch process."""
        self.popen.terminate()

    def close(self) -> None:
        """Terminates this roslaunch process."""
        self.terminate()



# -*- coding: utf-8 -*-
from loguru import logger as _logger
_logger.disable('roswire')

from .version import __version__
from .system import System, SystemDescription
from .roswire import ROSWire
from . import name



# -*- coding: utf-8 -*-
"""
This module provides a bag file writer that writes the contents of a bag to a
binary file on disk.

See
---
https://github.com/ros/ros_comm/blob/melodic-devel/tools/rosbag/src/rosbag/bag.py
"""
__all__ = ('BagWriter',)

from typing import BinaryIO, Iterable, Dict, Type, Tuple, List, Optional

from loguru import logger

from .core import (BagMessage, OpCode, Compression, ConnectionInfo, Chunk,
                   Index, IndexEntry, ChunkConnection)
from ..definitions import Message
from ..definitions.encode import (encode_uint32, write_uint32,
                                  encode_uint64,
                                  encode_time, write_time,
                                  write_encoded_header)

BIN_CHUNK_INFO_VERSION = encode_uint32(1)
BIN_INDEX_VERSION = encode_uint32(1)


class BagWriter:
    """
    Provides an interface for writing messages to a bag file on disk.

    Attributes
    ----------
    filename: str
        The name of the file to which the bag will be written.
    """
    def __init__(self, fn: str) -> None:
        self.__fn = fn
        self.__fp: BinaryIO = open(fn, 'wb')
        self.__connections: Dict[str, ConnectionInfo] = {}
        self.__chunks: List[Chunk] = []
        self.__pos_header = 0
        self.__pos_chunks = 0
        self.__pos_index = 0

    @property
    def filename(self) -> str:
        return self.__fn

    def _write_header(self,
                      code: Optional[OpCode],
                      fields: Dict[str, bytes]
                      ) -> None:
        if code:
            fields['op'] = code.value
        write_encoded_header(fields, self.__fp)

    def _write_header_record(self) -> None:
        logger.debug("writing bag header record")
        self.__fp.seek(self.__pos_header)
        self._write_header(OpCode.HEADER, {
            'chunk_count': encode_uint32(len(self.__chunks)),
            'conn_count': encode_uint32(len(self.__connections)),
            'index_pos': encode_uint64(self.__pos_index)
        })

        # ensure the bag header record is 4096 characters long by padding it
        # with ASCII space characters (0x20) where necessary.
        #
        # the ROS bag format doesn't consider size fields when computing the
        # length of the header and the record
        pos_current = self.__fp.tell()
        size = 4096
        size_header = pos_current - self.__pos_header - 4
        size_padding = size - size_header

        write_uint32(size_padding, self.__fp)
        padding = b'\x20' * size_padding
        self.__fp.write(padding)
        logger.debug("wrote bag header record")

    def _write_message(self,
                       pos_chunk_start: int,
                       index: Index,
                       message: BagMessage
                       ) -> None:
        typ = message.message.__class__
        connection = self._get_connection(message.topic, typ)
        logger.debug(f"writing message on connection [{connection.topic}]")

        pos_header = self.__fp.tell()
        self._write_header(OpCode.MESSAGE_DATA, {
            'conn': encode_uint32(connection.conn),
            'time': encode_time(message.time)})

        bin_data = message.message.encode()
        size_data = len(bin_data)
        write_uint32(size_data, self.__fp)
        self.__fp.write(bin_data)

        # update index
        offset = pos_header - pos_chunk_start
        index_entry = IndexEntry(time=message.time,
                                 pos=pos_header,
                                 offset=offset)
        if connection.conn not in index:
            index[connection.conn] = []
        index[connection.conn].append(index_entry)

    def _write_chunk_data(self, messages: Iterable[BagMessage]) -> Index:
        logger.debug("writing messages to chunk")
        index: Index = {}
        pos_start = self.__fp.tell()
        for m in messages:
            self._write_message(pos_start, index, m)
        logger.debug("wrote messages to chunk")
        return index

    def _write_chunk_record(self,
                            compression: Compression,
                            messages: Iterable[BagMessage]
                            ) -> Tuple[Chunk, Index]:
        bin_compression = compression.value.encode('utf-8')

        # for now, we write a bogus header and size field
        # once we've finished writing the data, we'll correct them
        pos_header = self.__fp.tell()
        self._write_header(OpCode.CHUNK, {
            'compression': bin_compression,
            'size': encode_uint32(0)})
        write_uint32(0, self.__fp)
        pos_data = self.__fp.tell()

        # write chunk contents
        index = self._write_chunk_data(messages)

        # determine time of earliest and latest message in the bag
        times = [e.time for ci in index.values() for e in ci]
        time_start = min(times)
        time_end = max(times)

        # compute chunk size
        pos_end = self.__fp.tell()
        size_compressed = pos_end - pos_data
        size_uncompressed = size_compressed

        # update header and size
        self.__fp.seek(pos_header)
        self._write_header(OpCode.CHUNK, {
            'compression': bin_compression,
            'size': encode_uint32(size_uncompressed)})
        write_uint32(size_compressed, self.__fp)
        self.__fp.seek(pos_end)

        # build a description of the chunk
        conns = [ChunkConnection(conn, len(entries))
                 for conn, entries in index.items()]
        chunk = Chunk(pos_record=pos_header,  # type: ignore
                      pos_data=pos_data,
                      time_start=time_start,
                      time_end=time_end,
                      connections=conns,
                      compression=compression,
                      size_compressed=size_compressed,
                      size_uncompressed=size_uncompressed)
        return chunk, index

    def _write_connection_index(self,
                                conn: int,
                                entries: List[IndexEntry]
                                ) -> None:
        logger.debug(f"writing connection index [{conn}]")
        num_entries = len(entries)
        size_data = num_entries * 12
        self._write_header(OpCode.INDEX_DATA, {
            'ver': BIN_INDEX_VERSION,
            'conn': encode_uint32(conn),
            'count': encode_uint32(num_entries)})
        write_uint32(size_data, self.__fp)
        for entry in entries:
            write_time(entry.time, self.__fp)
            write_uint32(entry.offset, self.__fp)

    def _write_chunk(self,
                     compression: Compression,
                     messages: Iterable[BagMessage]
                     ) -> None:
        # TODO for now, we only support uncompressed writing
        assert compression == Compression.NONE
        chunk, index = self._write_chunk_record(compression, messages)
        self.__chunks.append(chunk)
        for conn, entries in index.items():
            self._write_connection_index(conn, entries)

    def _write_connection_record(self, conn: ConnectionInfo) -> None:
        self._write_header(OpCode.CONNECTION_INFO, {
            'conn': encode_uint32(conn.conn),
            'topic': conn.topic.encode('utf-8')})

        # write the connection header
        header_conn: Dict[str, bytes] = {}
        if conn.callerid is not None:
            header_conn['callerid'] = conn.callerid.encode('utf-8')
        if conn.latching is not None:
            header_conn['latching'] = conn.latching.encode('utf-8')
        header_conn['topic'] = conn.topic_original.encode('utf-8')
        header_conn['type'] = conn.typ.encode('utf-8')
        header_conn['md5sum'] = conn.md5sum.encode('utf-8')
        header_conn['message_definition'] = \
            conn.message_definition.encode('utf-8')
        self._write_header(None, header_conn)

    def _get_connection(self,
                        topic: str,
                        typ: Type[Message]
                        ) -> ConnectionInfo:
        # if there isn't a connection for the topic, create one and write a
        # connection record.
        if topic not in self.__connections:
            msg_format = typ.format
            conn = len(self.__connections)
            info = ConnectionInfo(conn=conn,
                                  topic=topic,
                                  topic_original=topic,
                                  typ=msg_format.fullname,
                                  md5sum=typ.md5sum(),
                                  message_definition=msg_format.definition,
                                  callerid=None,
                                  latching=None)
            self.__connections[topic] = info
            self._write_connection_record(info)

        return self.__connections[topic]

    def _write_chunk_info_record(self, chunk: Chunk) -> None:
        num_connections = len(chunk.connections)
        # pos_header = self.__fp.tell()
        self._write_header(OpCode.CHUNK_INFO, {
            'ver': BIN_CHUNK_INFO_VERSION,
            'chunk_pos': encode_uint64(chunk.pos_record),
            'start_time': encode_time(chunk.time_start),
            'end_time': encode_time(chunk.time_end),
            'count': encode_uint32(num_connections)})

        size_data = num_connections * 8
        write_uint32(size_data, self.__fp)
        for connection in chunk.connections:
            write_uint32(connection.uid, self.__fp)
            write_uint32(connection.count, self.__fp)

    def _write_index(self) -> None:
        for connection in self.__connections.values():
            self._write_connection_record(connection)
        for chunk in self.__chunks:
            self._write_chunk_info_record(chunk)

    def write(self, messages: Iterable[BagMessage]) -> None:
        """
        Writes a sequence of messages to the bag.
        Any existing bag file contents will be overwritten.
        """
        self.__fp.truncate(0)
        self.__fp.write('#ROSBAG V2.0\n'.encode('utf-8'))

        # create a placeholder header for now
        self.__pos_header = self.__fp.tell()
        self._write_header_record()

        # for now, we write to a single, uncompressed chunk
        # each chunk record is followed by a sequence of IndexData record
        # - each connection in the chunk is represented by an IndexData record
        self.__pos_chunks = self.__fp.tell()
        self._write_chunk(Compression.NONE, messages)

        # write index
        self.__pos_index = self.__fp.tell()
        self._write_index()

        # fix the header
        self._write_header_record()

    def close(self) -> None:
        self.__fp.close()



# -*- coding: utf-8 -*-
__all__ = ('BagReader',)

from typing import (BinaryIO, Dict, Optional, Tuple, List, Type, Collection,
                    Set, Iterator)
from io import BytesIO
from functools import reduce
import os
import heapq

from loguru import logger

from .core import (OpCode, Compression, BagMessage, ChunkConnection, Chunk,
                   ConnectionInfo, BagHeader, IndexEntry, Index)
from ..definitions.base import Time, Duration
from ..definitions.msg import Message
from ..definitions.type_db import TypeDatabase
from ..definitions.decode import (decode_uint32, read_uint32,
                                  decode_uint64,
                                  decode_string,
                                  decode_time, read_time,
                                  read_sized,
                                  read_encoded_header)


class BagReader:
    def __init__(self, fn: str, db_type: TypeDatabase) -> None:
        self.__fp = open(fn, 'rb')
        self.__db_type = db_type
        self.__size_bytes: int = os.path.getsize(fn)

        version = self._read_version()
        assert version == '#ROSBAG V2.0'
        logger.debug(f"bag version: {version}")

        self.__header = self._read_header_record()
        logger.debug(f"bag header: {self.__header}")

        # skip past chunks
        self._seek(self.__header.index_pos)

        # obtain a list of all connections in the bag
        connections: List[ConnectionInfo] = []
        for i in range(self.__header.conn_count):
            conn = self._read_connection_record()
            connections.append(conn)
        self.__connections: Tuple[ConnectionInfo, ...] = tuple(connections)

        # obtain a summary of each chunk
        chunks: List[Chunk] = []
        for i in range(self.__header.chunk_count):
            info = self._read_chunk_info_record()
            chunks.append(info)
        self.__chunks: Tuple[Chunk, ...] = tuple(chunks)
        self.__pos_to_chunk: Dict[int, Chunk] = \
            {c.pos_record: c for c in chunks}

        # read the index
        self.__index: Index = self._read_index()
        logger.debug(f"topics: {self.topics}")
        for conn_id, indices in self.__index.items():
            topic = self.__connections[conn_id].topic
            logger.debug(f"conn {conn_id} [{topic}]: {len(indices)} messages")

    @property
    def connections(self) -> Tuple[ConnectionInfo, ...]:
        return self.__connections

    @property
    def header(self) -> BagHeader:
        return self.__header

    @property
    def chunks(self) -> Tuple[Chunk, ...]:
        return self.__chunks

    @property
    def index(self) -> Index:
        return self.__index

    @property
    def duration(self) -> Duration:
        """The duration of the recording in this bag file."""
        return Duration.between(self.time_start, self.time_end)

    @property
    def time_start(self) -> Time:
        """The time at which the recording began."""
        times = [c.time_start for c in self.chunks]
        return reduce(min, times[1:], times[0])

    @property
    def time_end(self) -> Time:
        """The time at which the recording ended."""
        times = [c.time_end for c in self.chunks]
        return reduce(max, times[1:], times[0])

    @property
    def topics(self) -> Set[str]:
        """The names of all topics represented in this bag."""
        return set(c.topic for c in self.connections)

    @property
    def topics_to_types(self) -> Dict[str, Type[Message]]:
        """A mapping from topics to their respective types."""
        return {c.topic: self.__db_type[c.typ] for c in self.connections}

    def _seek(self, pos: int, ptr: Optional[BinaryIO] = None) -> None:
        ptr = ptr if ptr else self.__fp
        ptr.seek(pos)

    def _skip_sized(self, ptr: Optional[BinaryIO] = None) -> None:
        ptr = ptr if ptr else self.__fp
        ptr.seek(read_uint32(ptr), os.SEEK_CUR)

    def _skip_record(self, ptr: Optional[BinaryIO] = None) -> None:
        self._skip_sized(ptr)
        self._skip_sized(ptr)

    def _read_version(self) -> str:
        return decode_string(self.__fp.readline()).rstrip()

    def _read_header(self,
                     op_expected: Optional[OpCode] = None,
                     *,
                     ptr: Optional[BinaryIO] = None
                     ) -> Dict[str, bytes]:
        fields = read_encoded_header(ptr if ptr else self.__fp)
        if op_expected:
            assert 'op' in fields
            op_actual: OpCode = OpCode(fields['op'])
            if fields['op'] != op_expected.value:
                m = ("unexpected opcode when reading header: "
                     f"expected {op_expected.name} [{op_expected.hex}] "
                     f"but was {op_actual.name} [{op_actual.hex}].")
                raise Exception(m)
        return fields

    def _read_header_record(self) -> BagHeader:
        header = self._read_header(OpCode.HEADER)
        index_pos = decode_uint64(header['index_pos'])
        conn_count = decode_uint32(header['conn_count'])
        chunk_count = decode_uint32(header['chunk_count'])
        self._skip_sized()
        return BagHeader(index_pos, conn_count, chunk_count)

    def _read_connection_record(self) -> ConnectionInfo:
        header = self._read_header(OpCode.CONNECTION_INFO)
        conn = self._read_header()
        logger.debug(f"conn record header: {header}")
        logger.debug(f"conn header: {conn}")
        callerid: Optional[str] = None
        latching: Optional[str] = None
        if 'callerid' in conn:
            callerid = decode_string(conn['callerid'])
        if 'latching' in conn:
            latching = decode_string(conn['latching'])
        return ConnectionInfo(conn=decode_uint32(header['conn']),
                              callerid=callerid,
                              latching=latching,
                              topic=decode_string(header['topic']),
                              topic_original=decode_string(conn['topic']),
                              typ=decode_string(conn['type']),
                              md5sum=decode_string(conn['md5sum']),
                              message_definition=decode_string(conn['message_definition']))  # noqa

    def _read_chunk_info_record(self) -> Chunk:
        header = self._read_header(OpCode.CHUNK_INFO)
        ver: int = decode_uint32(header['ver'])
        assert ver == 1
        pos_record: int = decode_uint64(header['chunk_pos'])
        time_start: Time = decode_time(header['start_time'])
        time_end: Time = decode_time(header['end_time'])
        num_connections: int = decode_uint32(header['count'])

        # obtain a summary of the number of messages for each connection
        # represented in this chunk
        connections: List[ChunkConnection] = []
        contents: bytes = read_sized(self.__fp)
        for i in range(num_connections):
            uid = decode_uint32(contents[0:4])
            count = decode_uint32(contents[4:8])
            connections.append(ChunkConnection(uid, count))
            contents = contents[8:]

        # read connection index records?
        assert not contents

        # read the chunk header
        pos_original = self.__fp.tell()
        self.__fp.seek(pos_record)
        header = self._read_header(OpCode.CHUNK)
        size_uncompressed = decode_uint32(header['size'])
        compression = Compression(decode_string(header['compression']))
        pos_data = self.__fp.tell()

        # determine the compressed size of the chunk data
        size_compressed = read_uint32(self.__fp)

        # restore the original position of the read pointer
        self.__fp.seek(pos_original)

        chunk = Chunk(pos_record=pos_record,
                      pos_data=pos_data,
                      time_start=time_start,
                      time_end=time_end,
                      size_uncompressed=size_uncompressed,
                      size_compressed=size_compressed,
                      compression=compression,
                      connections=connections)

        logger.debug("decoded chunk: {chunk}")
        return chunk

    def _read_index(self) -> Index:
        logger.debug("reading index")
        index: Index = {c.conn: [] for c in self.connections}
        for chunk in self.chunks:
            logger.debug(f"reading index for chunk: {chunk}")
            pos = chunk.pos_record
            self._seek(pos)
            self._skip_record()
            for i in range(len(chunk.connections)):
                self._read_index_record(pos, index)
        logger.debug("read index")

        logger.debug("pruning empty connections from index")
        for c in [cc for cc in index if not index[cc]]:
            del index[c]
            # TODO delete connection!
        logger.debug("pruned index")
        return index

    def _read_index_record(self, pos_chunk: int, index: Index) -> None:
        header = self._read_header(OpCode.INDEX_DATA)
        ver = decode_uint32(header['ver'])
        uid = decode_uint32(header['conn'])
        count = decode_uint32(header['count'])
        assert ver == 1

        read_uint32(self.__fp)  # skip size
        for i in range(count):
            time = read_time(self.__fp)
            offset = read_uint32(self.__fp)
            entry = IndexEntry(time=time, pos=pos_chunk, offset=offset)
            index[uid].append(entry)

    def _get_connections(self,
                         topics: Optional[Collection[str]] = None
                         ) -> Iterator[ConnectionInfo]:
        """
        Returns an iterator over all connections for a given set of topics.
        """
        if not topics:
            yield from self.connections
        else:
            yield from (c for c in self.connections if c.topic in topics)

    def _get_entries(self,
                     connections: Collection[ConnectionInfo],
                     time_start: Optional[Time] = None,
                     time_end: Optional[Time] = None
                     ) -> Iterator[IndexEntry]:
        entries = heapq.merge(*[self.__index[c.conn] for c in connections])
        for entry in entries:
            if time_start and entry.time < time_start:
                continue
            if time_end and entry.time > time_end:
                return
            yield entry

    def fetch_message_data_record(self, pos: int, offset: int) -> BagMessage:
        # find the chunk to which the message belongs
        # read the contents of that chunk to a bytes buffer
        chunk = self.__pos_to_chunk[pos]
        self._seek(chunk.pos_data)
        if chunk.compression == Compression.NONE:
            bfr = BytesIO(read_sized(self.__fp))
        else:
            raise NotImplementedError

        # seek position of message data record
        # - skip any preceding connection records
        self._seek(offset, bfr)
        while True:
            header = self._read_header(ptr=bfr)
            op = OpCode(header['op'])
            if op == OpCode.CONNECTION_INFO:
                self._skip_sized(bfr)
                continue
            if op == OpCode.MESSAGE_DATA:
                conn_id = decode_uint32(header['conn'])
                t = decode_time(header['time'])
                break
            m = "unexpected opcode: got {} but expected {}"
            m = m.format(op, OpCode.MESSAGE_DATA)
            raise Exception(m)

        # fetch the topic name and message type
        conn_info = self.__connections[conn_id]
        topic = conn_info.topic
        msg_typ_name = conn_info.typ
        msg_typ = self.__db_type[msg_typ_name]

        # read the raw message data
        raw = read_sized(bfr)
        content = msg_typ.read(BytesIO(raw))
        msg = BagMessage(topic, t, content)
        logger.debug(f"decoded message: {msg}")
        return msg

    def read_messages(self,
                      topics: Optional[Collection[str]] = None,
                      time_start: Optional[Time] = None,
                      time_end: Optional[Time] = None
                      ) -> Iterator[BagMessage]:
        conns = set(self._get_connections(topics))
        for entry in self._get_entries(conns, time_start, time_end):
            yield self.fetch_message_data_record(entry.pos, entry.offset)

    def __iter__(self) -> Iterator[BagMessage]:
        yield from self.read_messages()



# -*- coding: utf-8 -*-
"""
This file implements a proxy for parsing the contents of launch files.
"""
__all__ = ('LaunchFileReader',)

from typing import (Any, Callable, Collection, Optional, overload, Sequence,
                    Tuple, Union)
import xml.etree.ElementTree as ET

from loguru import logger
import attr
import dockerblade

from .rosparam import load_from_yaml_string as load_rosparam_from_string
from .config import LaunchConfig, NodeConfig
from .context import LaunchContext
from .substitution import ArgumentResolver
from ...name import (namespace_join, global_name, namespace, name_is_global,
                     name_is_private)
from ...exceptions import FailedToParseLaunchFile

_TAG_TO_LOADER = {}

Loader = \
    Callable[['LaunchFileReader', LaunchContext, LaunchConfig, ET.Element],
             Tuple[LaunchContext, LaunchConfig]]


def _read_contents(tag: ET.Element) -> str:
    """Reads the text contents of an XML element."""
    # FIXME add support for CDATA -- possibly via lxml or xml.dom?
    return ''.join(t.text for t in tag if t.text)


def _parse_bool(attr: str, val: str) -> bool:
    """Parses a boolean value from an XML attribute."""
    val = val.lower()
    if val == 'true':
        return True
    if val == 'false':
        return False

    m = f'illegal boolean attribute [{attr}]: {val}'
    raise FailedToParseLaunchFile(m)


def _parse_float(attr: str, val: str) -> float:
    """Parses a float value from an XML attribute."""
    if not val:
        m = f'empty string used by float attribute [{attr}]'
        raise FailedToParseLaunchFile(m)
    try:
        return float(val)
    except ValueError:
        m = f'failed to parse attribute [{attr}] to float: {val}'
        raise FailedToParseLaunchFile(m)


def convert_str_to_type(s: str, typ: str) -> Union[bool, int, str, float]:
    if typ == 'auto':
        if s.lower() in ('true', 'false'):
            return convert_str_to_type(s, 'bool')
        if s.isnumeric() and '.' in s:
            return convert_str_to_type(s, 'float')
        if s.isnumeric():
            return convert_str_to_type(s, 'int')
        return convert_str_to_type(s, 'str')

    if typ in ('str', 'string'):
        return s
    if typ in ('float', 'double'):
        return float(s)
    if typ == 'int':
        return int(s)
    if typ in ('bool', 'boolean'):
        s = s.lower()
        if s in ('true', '1'):
            return True
        if s in ('false', '0'):
            return False
        raise ValueError(f"illegal 'bool' value: {s}")

    raise ValueError(f'unknown parameter type: {typ}')


def tag(name: str,
        legal_attributes: Collection[str] = tuple()
        ) -> Callable[[Loader], Loader]:
    legal_attributes = frozenset(list(legal_attributes) + ['if', 'unless'])

    def wrap(loader: Loader) -> Loader:
        def wrapped(self: 'LaunchFileReader',
                    ctx: LaunchContext,
                    cfg: LaunchConfig,
                    elem: ET.Element
                    ) -> Tuple[LaunchContext, LaunchConfig]:
            logger.debug(f"parsing <{name}> tag")
            for attribute in elem.attrib:
                if attribute not in legal_attributes:
                    m = f'<{name}> tag contains illegal attribute: {attribute}'
                    raise FailedToParseLaunchFile(m)

            # should we process this element?
            if not self._ifunless_check(elem, ctx):
                return ctx, cfg
            return loader(self, ctx, cfg, elem)

        _TAG_TO_LOADER[name] = wrapped
        return wrapped

    return wrap


@attr.s(auto_attribs=True)
class LaunchFileReader:
    _shell: dockerblade.Shell
    _files: dockerblade.FileSystem

    def _parse_file(self, fn: str) -> ET.Element:
        """Parses a given XML launch file to a root XML element."""
        contents = self._files.read(fn)
        logger.info(f'parsing launch file [{fn}]:\n{contents}')
        root = ET.fromstring(contents)
        if root.tag != 'launch':
            m = 'root of launch file must have <launch></launch> tags'
            raise FailedToParseLaunchFile(m)
        return root

    def _load_tags(self,
                   ctx: LaunchContext,
                   cfg: LaunchConfig,
                   tags: Sequence[ET.Element]
                   ) -> Tuple[LaunchContext, LaunchConfig]:
        for tag in (t for t in tags if t.tag in _TAG_TO_LOADER):
            loader = _TAG_TO_LOADER[tag.tag]
            ctx, cfg = loader(self, ctx, cfg, tag)
        return ctx, cfg

    @tag('group', ['ns'])
    def _load_group_tag(self,
                        ctx: LaunchContext,
                        cfg: LaunchConfig,
                        tag: ET.Element
                        ) -> Tuple[LaunchContext, LaunchConfig]:
        # create context for group
        ns = self._read_namespace(ctx, tag)
        ctx_child = ctx.child(ns)

        # handle nested tags
        nested_tags = [t for t in tag]
        ctx_child, cfg = self._load_tags(ctx_child, cfg, nested_tags)

        return ctx, cfg

    @tag('param', ['name', 'value', 'type', 'textfile', 'binfile', 'command'])
    def _load_param_tag(self,
                        ctx: LaunchContext,
                        cfg: LaunchConfig,
                        tag: ET.Element
                        ) -> Tuple[LaunchContext, LaunchConfig]:
        name = self._read_required(tag, 'name', ctx)
        typ = self._read_optional(tag, 'type', ctx) or 'auto'
        logger.debug(f"adding parameter [{name}] with type [{typ}]")

        # obtain value from either 'value', 'textfile', 'binfile', or
        # 'command' exclusively.
        val = self._read_optional(tag, 'value', ctx)
        textfile = self._read_optional(tag, 'textfile', ctx)
        binfile = self._read_optional(tag, 'binfile', ctx)
        command = self._read_optional(tag, 'command', ctx)
        value_candidates = [val, textfile, binfile, command]

        if len([c for c in value_candidates if c is not None]) != 1:
            m = '<param> must have exactly one value/textfile/binfile/command'
            raise FailedToParseLaunchFile(m)

        value: Any
        if val is not None:
            value = convert_str_to_type(val, typ)
        if textfile is not None:
            value = self._files.read(textfile)
        if binfile is not None:
            value = self._files.read(binfile, binary=True)
        if command is not None:
            logger.debug(f'obtaining value for parameter [{name}] '
                         f'via command:\n{command}')
            value = self._shell.check_output(command, text=True, stderr=False)

        logger.debug(f"obtained value for parameter [{name}]: {value}")

        # compute the fully qualified name
        if name_is_global(name):
            fullname = name
        elif name_is_private(name):
            fullname = namespace_join(ctx.namespace, name[1:])
        else:
            fullname = namespace_join(ctx.namespace, name)

        # register the parameter
        cfg = cfg.with_param(name=fullname,
                             typ=typ,
                             value=value,
                             command=command)

        return ctx, cfg

    @tag('rosparam', ['command', 'ns', 'file', 'param', 'subst_value'])
    def _load_rosparam_tag(self,
                           ctx: LaunchContext,
                           cfg: LaunchConfig,
                           tag: ET.Element
                           ) -> Tuple[LaunchContext, LaunchConfig]:
        filename = self._read_optional(tag, 'file', ctx)
        subst_value = self._read_optional_bool(tag, 'subst_value', ctx, False)
        ns = self._read_optional(tag, 'ns', ctx) or ''
        param = self._read_optional(tag, 'param', ctx) or ''
        param = namespace_join(ns, param)
        full_param = namespace_join(ctx.namespace, param)
        value = _read_contents(tag)

        cmd: str = self._read_optional(tag, 'command', ctx) or 'load'
        if cmd not in ('load', 'delete', 'dump'):
            m = f"<rosparam> unsupported 'command': {cmd}"
            raise FailedToParseLaunchFile(m)

        if cmd == 'load' and filename is not None:
            if not self._files.isfile(filename):
                m = f"<rosparam> file does not exist: {filename}"
                raise FailedToParseLaunchFile(m)

        if cmd == 'delete' and filename is not None:
            m = "<rosparam> command:delete does not support filename"
            raise FailedToParseLaunchFile(m)

        # handle load command
        if cmd == 'load':
            if filename is None:
                yml_text = value
            else:
                yml_text = self._files.read(filename)

            if subst_value:
                yml_text = self._resolve_args(yml_text, ctx)
            logger.debug(f"parsing rosparam YAML:\n{yml_text}")
            data = load_rosparam_from_string(yml_text)
            logger.debug(f"rosparam values: {data}")
            if not isinstance(data, dict) and not param:
                m = "<rosparam> requires 'param' for non-dictionary values"
                raise FailedToParseLaunchFile(m)
            cfg = cfg.with_rosparam(name=full_param, value=data)

        # handle dump command
        if cmd == 'dump':
            m = "'dump' command is currently not supported in <rosparam>"
            raise NotImplementedError(m)

        # handle delete command
        if cmd == 'delete':
            m = "'delete' command is currently not supported in <rosparam>"
            raise NotImplementedError(m)

        return ctx, cfg

    @tag('remap', ['from', 'to'])
    def _load_remap_tag(self,
                        ctx: LaunchContext,
                        cfg: LaunchConfig,
                        tag: ET.Element
                        ) -> Tuple[LaunchContext, LaunchConfig]:
        frm = self._read_required(tag, 'from', ctx)
        to = self._read_required(tag, 'to', ctx)
        ctx = ctx.with_remapping(frm, to)
        return ctx, cfg

    @tag('node', ['name', 'type', 'pkg', 'required', 'clear_params',
                  'respawn', 'ns', 'output', 'args', 'ns', 'launch-prefix'])
    def _load_node_tag(self,
                       ctx: LaunchContext,
                       cfg: LaunchConfig,
                       tag: ET.Element
                       ) -> Tuple[LaunchContext, LaunchConfig]:
        name = self._read_required(tag, 'name', ctx)
        package = self._read_required(tag, 'pkg', ctx)
        node_type = self._read_required(tag, 'type', ctx)
        output = self._read_optional(tag, 'output', ctx)
        launch_prefix = self._read_optional(tag, 'launch-prefix', ctx)
        cwd = self._read_optional(tag, 'cwd', ctx)
        args = self._read_optional(tag, 'args', ctx)
        required = self._read_optional_bool(tag, 'required', ctx, False)
        respawn = self._read_optional_bool(tag, 'respawn', ctx, False)
        respawn_delay = \
            self._read_optional_float(tag, 'respawn_delay', ctx, 0.0)

        # create context
        ns = self._read_namespace(ctx, tag)
        ctx_child = ctx.node_child(ns, name)

        # handle 'clear_params'
        if self._read_optional_bool(tag, 'clear_params', ctx, False):
            clear_ns = global_name(ctx_child.namespace)
            cfg = cfg.with_clear_param(clear_ns)

        # handle nested tags
        allowed = {'env', 'remap', 'param', 'rosparam'}
        nested_tags = [t for t in tag if t.tag in allowed]
        ctx_child, cfg = self._load_tags(ctx_child, cfg, nested_tags)

        node = NodeConfig(name=name,
                          namespace=namespace(ctx_child.namespace),
                          package=package,
                          cwd=cwd,
                          args=args,
                          required=required,
                          respawn=respawn,
                          respawn_delay=respawn_delay,
                          output=output,
                          remappings=ctx_child.remappings,
                          launch_prefix=launch_prefix,
                          filename=ctx.filename,
                          env_args=ctx_child.env_args,
                          typ=node_type)
        cfg = cfg.with_node(node)
        return ctx, cfg

    @tag('arg', ['name', 'default', 'value', 'doc'])
    def _load_arg_tag(self,
                      ctx: LaunchContext,
                      cfg: LaunchConfig,
                      tag: ET.Element
                      ) -> Tuple[LaunchContext, LaunchConfig]:
        name = self._read_required(tag, 'name', ctx)
        value = self._read_optional(tag, 'value', ctx)
        default = self._read_optional(tag, 'default', ctx)
        doc = self._read_optional(tag, 'doc', ctx)
        ctx = ctx.with_arg(name=name,
                           value=value,
                           default=default,
                           doc=doc)
        return ctx, cfg

    @tag('env', ['name', 'value'])
    def _load_env_tag(self,
                      ctx: LaunchContext,
                      cfg: LaunchConfig,
                      tag: ET.Element
                      ) -> Tuple[LaunchContext, LaunchConfig]:
        name = self._read_required(tag, 'name', ctx)
        value = self._read_required(tag, 'value', ctx)
        cfg = cfg.with_env(name, value)
        ctx = ctx.with_env_arg(name, value)
        return ctx, cfg

    @tag('include', ['file', 'pass_all_args', 'ns', 'clear_params'])
    def _load_include_tag(self,
                          ctx: LaunchContext,
                          cfg: LaunchConfig,
                          tag: ET.Element
                          ) -> Tuple[LaunchContext, LaunchConfig]:
        include_filename = self._read_required(tag, 'file', ctx)
        logger.debug(f"include file: {include_filename}")
        cfg = cfg.with_roslaunch_file(include_filename)

        # create context
        ns = self._read_namespace(ctx, tag)
        ctx_child = ctx.include_child(ns, include_filename)

        # handle 'clear_params'
        if self._read_optional_bool(tag, 'clear_params', ctx, False):
            if not ns:
                m = "'ns' must be specified to use 'clear_params'"
                raise FailedToParseLaunchFile(m)
            cfg = cfg.with_clear_param(ctx_child.namespace)

        # if instructed to pass along args, then those args must be added to
        # the child context
        if self._read_optional_bool(tag, 'pass_all_args', ctx, False):
            ctx_child = ctx_child.with_pass_all_args()

        # handle child tags
        child_tags = [t for t in tag if t.tag in ('env', 'arg')]
        ctx_child, cfg = self._load_tags(ctx_child, cfg, child_tags)
        ctx_child = ctx_child.process_include_args()
        logger.debug(f"prepared include context: {ctx_child}")

        logger.debug("loading include file")
        launch = self._parse_file(include_filename)
        ctx_child, cfg = self._load_tags(ctx_child, cfg, list(launch))

        return ctx, cfg

    def _read_namespace(self,
                        ctx: LaunchContext,
                        tag: ET.Element
                        ) -> Optional[str]:
        """Fetches the value of the optional 'ns' attribute."""
        ns = self._read_optional(tag, 'ns', ctx)
        if ns == '':
            m = f"<{tag.tag}> has empty attribute [ns]"
            raise FailedToParseLaunchFile(m)
        return ns

    @overload
    def _read_optional_bool(self,
                            elem: ET.Element,
                            attrib: str,
                            ctx: LaunchContext,
                            default: None
                            ) -> Optional[bool]: ...

    @overload
    def _read_optional_bool(self,
                            elem: ET.Element,
                            attrib: str,
                            ctx: LaunchContext,
                            default: bool
                            ) -> bool: ...

    def _read_optional_bool(self,
                            elem: ET.Element,
                            attrib: str,
                            ctx: LaunchContext,
                            default: Optional[bool] = None
                            ) -> Optional[bool]:
        s = self._read_optional(elem, attrib, ctx)
        if s is None:
            return default
        return _parse_bool(attrib, s)

    @overload
    def _read_optional_float(self,
                             elem: ET.Element,
                             attrib: str,
                             ctx: LaunchContext,
                             default: None
                             ) -> Optional[float]: ...

    @overload
    def _read_optional_float(self,
                             elem: ET.Element,
                             attrib: str,
                             ctx: LaunchContext,
                             default: float
                             ) -> float: ...

    def _read_optional_float(self,
                             elem: ET.Element,
                             attrib: str,
                             ctx: LaunchContext,
                             default: Optional[float] = None
                             ) -> Optional[float]:
        s = self._read_optional(elem, attrib, ctx)
        if s is None:
            return default
        return _parse_float(attrib, s)

    def _read_optional(self,
                       elem: ET.Element,
                       attrib: str,
                       ctx: LaunchContext
                       ) -> Optional[str]:
        """Reads the string value of an optional attribute of an element."""
        if attrib not in elem.attrib:
            return None
        return self._read_required(elem, attrib, ctx)

    def _read_required(self,
                       elem: ET.Element,
                       attrib: str,
                       ctx: LaunchContext
                       ) -> str:
        """Reads the string value of a required attribute of an element."""
        return self._resolve_args(elem.attrib[attrib], ctx)

    def _ifunless_check(self, elem: ET.Element, ctx: LaunchContext) -> bool:
        """Determines whether an element should be parsed."""
        if_val = self._read_optional(elem, 'if', ctx)
        unless_val = self._read_optional(elem, 'unless', ctx)
        if if_val is not None and unless_val is not None:
            m = 'tag may not provide both "if" and "unless" attributes'
            raise FailedToParseLaunchFile(m)
        if if_val is not None:
            return _parse_bool('if', if_val)
        if unless_val is not None:
            return not _parse_bool('unless', unless_val)
        return True

    def _resolve_args(self, s: str, ctx: LaunchContext) -> str:
        """Resolves all substitution args in a given string."""
        logger.debug(f"resolve [{s}] with context: {ctx.resolve_dict}")
        resolve_ctx = ctx.resolve_dict or {}
        resolver = ArgumentResolver(shell=self._shell,
                                    files=self._files,
                                    context=resolve_ctx)
        return resolver.resolve(s)

    def read(self,
             fn: str,
             argv: Optional[Sequence[str]] = None
             ) -> LaunchConfig:
        """Parses the contents of a given launch file.

        Returns
        -------
        LaunchConfig
            A description of the launch configuration.

        Reference
        ---------
        * http://wiki.ros.org/roslaunch/XML/node
        * http://docs.ros.org/kinetic/api/roslaunch/html/roslaunch.xmlloader.XmlLoader-class.html
        """
        cfg = LaunchConfig()
        ctx = LaunchContext(namespace='/', filename=fn)
        if argv:
            ctx = ctx.with_argv(argv)

        launch = self._parse_file(fn)
        ctx, cfg = self._load_tags(ctx, cfg, list(launch))
        logger.debug(f"launch configuration: {cfg}")
        return cfg



# -*- coding: utf-8 -*-
"""
This file provides utilities for interacting with rosparam.
"""
__all__ = ('load_from_yaml_string',)

from typing import Dict, Any
import math
import re

import yaml


class YAMLLoader(yaml.SafeLoader):
    """A custom YAML loader for rosparam files."""


def load_from_yaml_string(s: str) -> Dict[str, Any]:
    """Parses the contents of a rosparam file to a dictionary."""
    return yaml.load(s, Loader=YAMLLoader) or {}


def __load_radians(loader: YAMLLoader, node: yaml.YAMLObject) -> float:
    """Safely converts rad(num) to a float value.

    Note
    ----
    This does not support evaluation of expressions.
    """
    expr_s = loader.construct_scalar(node).strip()
    if expr_s.startswith('rad('):
        expr_s = expr_s[4:-1]

    # TODO safely parse and evaluate expression
    return float(expr_s)


def __load_degrees(loader: YAMLLoader, node: yaml.YAMLObject) -> float:
    """Safely converts deg(num) to a float value."""
    expr_s = loader.construct_scalar(node).strip()
    if expr_s.startswith('def('):
        expr_s = expr_s[4:-1]
    return float(expr_s) * math.pi / 180.0


YAMLLoader.add_constructor('!degrees', __load_degrees)
YAMLLoader.add_implicit_resolver(
    '!degrees', re.compile('^deg\([^\)]*\)$'), first='deg(')
YAMLLoader.add_constructor('!radians', __load_radians)
YAMLLoader.add_implicit_resolver(
    '!radians', re.compile('^rad\([^\)]*\)$'), first='rad(')



# -*- coding: utf-8 -*-
"""
This module provides data structures that represent ROS launch configurations.
"""
__all__ = ('LaunchContext',)

from typing import Tuple, Any, Optional, Sequence, Dict
from copy import deepcopy

from loguru import logger
import attr

from ...name import canonical_name, name_is_legal, namespace_join
from ...exceptions import FailedToParseLaunchFile


@attr.s(frozen=True, slots=True)
class LaunchContext:
    filename: str = attr.ib()
    resolve_dict: Dict[str, Any] = attr.ib(factory=dict)
    parent: Optional['LaunchContext'] = attr.ib(default=None)
    namespace: str = attr.ib(default='/')
    arg_names: Tuple[str, ...] = attr.ib(default=tuple())
    env_args: Tuple[Tuple[str, str], ...] = attr.ib(default=tuple())
    pass_all_args: bool = attr.ib(default=False)
    include_resolve_dict: Optional[Dict[str, Any]] = attr.ib(default=None)
    remappings: Tuple[Tuple[str, str], ...] = attr.ib(default=tuple())
    node_name: Optional[str] = attr.ib(default=None)

    @property
    def is_node_context(self) -> bool:
        """Determines whether or not this context is node-local."""
        return self.node_name is not None

    def include_child(self,
                      ns: Optional[str],
                      filename: str
                      ) -> 'LaunchContext':
        ctx = self.child(ns)
        ctx = attr.evolve(ctx,
                          filename=filename,
                          arg_names=tuple(),
                          include_resolve_dict={})
        return ctx

    def node_child(self, ns: Optional[str], name: str) -> 'LaunchContext':
        ctx = self.child(ns).child(name)
        ctx = attr.evolve(ctx,
                          node_name=name,
                          include_resolve_dict={})
        return ctx

    def child(self, ns: Optional[str] = None) -> 'LaunchContext':
        """Creates a child context that inherits from this context."""
        if ns is None:
            child_ns = self.namespace
        elif ns.startswith('/') or ns == '~':
            child_ns = ns
        else:
            child_ns = namespace_join(self.namespace, ns)
        return attr.evolve(self,
                           parent=self,
                           namespace=child_ns,
                           pass_all_args=False)

    def with_remapping(self, frm: str, to: str) -> 'LaunchContext':
        """Adds a name remapping."""
        frm = canonical_name(frm)
        to = canonical_name(to)
        if not frm or not to:
            m = "'from' and 'to' attributes must be specified for <remap>"
            raise FailedToParseLaunchFile(m)
        if not name_is_legal(frm):
            m = f"<remap>: invalid ROS name [from]: {frm}"
            raise FailedToParseLaunchFile(m)
        if not name_is_legal(to):
            m = f"<remap>: invalid ROS name [to]: {to}"
            raise FailedToParseLaunchFile(m)

        # overwrite any existing remapping from the given source before adding
        # the given remapping
        remappings = tuple(r for r in self.remappings if r[0] != frm)
        remappings = remappings + ((frm, to), )
        return attr.evolve(self, remappings=remappings)

    def with_pass_all_args(self) -> 'LaunchContext':
        ctx = self
        if self.parent and 'arg' in self.parent.resolve_dict:
            for var, val in self.parent.resolve_dict['arg'].items():
                ctx = ctx.with_arg(var, value=val)
        return attr.evolve(ctx, pass_all_args=True)

    def process_include_args(self) -> 'LaunchContext':
        if self.include_resolve_dict is None:
            return attr.evolve(self, arg_names=tuple())

        arg_dict = self.include_resolve_dict.get('arg', {})
        for arg in self.arg_names:
            if arg not in arg_dict:
                m = f'include arg [{arg}] is missing value.'
                raise FailedToParseLaunchFile(m)

        return attr.evolve(self,
                           arg_names=tuple(),
                           resolve_dict=deepcopy(self.include_resolve_dict),
                           include_resolve_dict=None)

    def with_argv(self, argv: Sequence[str]) -> 'LaunchContext':
        # ignore parameter assignment mappings
        logger.debug(f"loading argv: {argv}")
        mappings: Dict[str, str] = {}
        for arg in (a for a in argv if ':=' in a):
            var, sep, val = [a.strip() for a in arg.partition(':=')]
            if not var.startswith('__'):
                mappings[var] = val
        logger.debug(f"loaded argv: {mappings}")
        resolve_dict = self.resolve_dict.copy()
        resolve_dict['arg'] = mappings
        return attr.evolve(self, resolve_dict=resolve_dict)

    def with_env_arg(self, var: str, val: Any) -> 'LaunchContext':
        env_args = self.env_args + ((var, val),)
        return attr.evolve(self, env_args=env_args)

    def with_arg(self,
                 name: str,
                 default: Optional[Any] = None,
                 value: Optional[Any] = None,
                 doc: Optional[str] = None
                 ) -> 'LaunchContext':
        logger.debug(f"adding arg [{name}] to context")
        arg_names = self.arg_names
        if name in self.arg_names:
            if not self.pass_all_args:
                m = f"arg [{name}] has already been declared"
                raise FailedToParseLaunchFile(m)
        else:
            arg_names = arg_names + (name,)

        # decide which resolve dictionary should be used
        use_include_resolve_dict = self.include_resolve_dict is not None
        if use_include_resolve_dict:
            assert self.include_resolve_dict is not None  # stupid mypy
            resolve_dict = self.include_resolve_dict.copy()
        else:
            resolve_dict = self.resolve_dict.copy()

        arg_dict = resolve_dict['arg'] = resolve_dict.get('arg', {}).copy()

        if value is not None:
            if name in arg_dict and not self.pass_all_args:
                m = f"arg [{name}] value has already been defined."
                raise FailedToParseLaunchFile(m)
            arg_dict[name] = value
        elif default is not None:
            arg_dict[name] = arg_dict.get(name, default)

        # construct new context
        ctx = attr.evolve(self, arg_names=arg_names)
        if use_include_resolve_dict:
            return attr.evolve(ctx, include_resolve_dict=resolve_dict)
        else:
            return attr.evolve(ctx, resolve_dict=resolve_dict)



# -*- coding: utf-8 -*-
__all__ = ('Env',)

import xml.etree.ElementTree as ET

import attr


@attr.s(frozen=True, slots=True, auto_attribs=True)
class Env:
    """Provides an environment variable definition.

    Attributes
    ----------
    name: str
        The name of the environment variable.
    value: str
        The concrete value of the environment variable.
    """
    name: str
    value: str

    def to_xml_element(self) -> ET.Element:
        element = ET.Element('env')
        element.attrib['name'] = self.name
        element.attrib['value'] = self.value
        return element



# -*- coding: utf-8 -*-
from .bag import BagRecorder, BagPlayer
from .catkin import CatkinInterface, CatkinTools, CatkinMake
from .container import Container, ContainerManager
from .node import Node, NodeManager
from .parameters import ParameterServer
from .roscore import ROSCore
from .service import Service, ServiceManager
from .roslaunch import ROSLaunchManager
from .state import SystemState



# -*- coding: utf-8 -*-
import pytest

from roswire.definitions import MsgFormat, SrvFormat, Package, PackageDatabase


def test_to_and_from_dict():
    pkg = 'tf'
    msg_tf = MsgFormat.from_dict({
        'package': pkg,
        'name': 'tfMessage',
        'definition': 'geometry_msgs/TransformStamped[] transforms\n',
        'fields': [{'type': 'geometry_msgs/TransformStamped[]',
                    'name': 'transforms'}]})
    srv_fg = SrvFormat.from_dict({
        'package': pkg,
        'name': 'FrameGraph',
        'definition': '---\nstring dot_graph\n',
        'response': {
            'definition': 'string dot_graph\n',
            'fields': [{'type': 'string', 'name': 'dot_graph'}]}})
    p = Package(name=pkg,
                path='/ros_ws/src/geometry/tf',
                messages=[msg_tf],
                actions=[],
                services=[srv_fg])
    assert p == Package.from_dict(p.to_dict())


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_build(filesystem):
    path = '/opt/ros/melodic/share/tf'
    expected = Package.from_dict({
        'path': path,
        'name': 'tf',
        'messages': [
            {'name': 'tfMessage',
             'definition': 'geometry_msgs/TransformStamped[] transforms\n',
             'fields': [{'type': 'geometry_msgs/TransformStamped[]',
                         'name': 'transforms'}]}
        ],
        'services': [
            {'name': 'FrameGraph',
             'definition': '---\nstring dot_graph\n',
             'response': {
                'definition': 'string dot_graph',
                'fields': [{'type': 'string',
                            'name': 'dot_graph'}]}}
        ]
    })
    actual = Package.build(path, filesystem)
    assert actual == expected


@pytest.mark.parametrize('sut', ['fetch'], indirect=True)
def test_database_paths(sut):
    expected = {
        '/opt/ros/melodic/share/moveit_ros_occupancy_map_monitor',
        '/opt/ros/melodic/share/common_msgs',
        '/opt/ros/melodic/share/nodelet_core',
        '/opt/ros/melodic/share/ros_comm',
        '/opt/ros/melodic/share/bond_core',
        '/opt/ros/melodic/share/ros_base',
        '/opt/ros/melodic/share/ros_core',
        '/opt/ros/melodic/share/roscpp_core',
        '/opt/ros/melodic/share/ros',
        '/opt/ros/melodic/share/actionlib',
        '/opt/ros/melodic/share/actionlib_msgs',
        '/opt/ros/melodic/share/amcl',
        '/opt/ros/melodic/share/angles',
        '/opt/ros/melodic/share/base_local_planner',
        '/opt/ros/melodic/share/bond',
        '/opt/ros/melodic/share/bondcpp',
        '/opt/ros/melodic/share/bondpy',
        '/opt/ros/melodic/share/camera_calibration_parsers',
        '/opt/ros/melodic/share/camera_info_manager',
        '/opt/ros/melodic/share/catkin',
        '/opt/ros/melodic/share/class_loader',
        '/opt/ros/melodic/share/clear_costmap_recovery',
        '/opt/ros/melodic/share/cmake_modules',
        '/opt/ros/melodic/share/control_msgs',
        '/opt/ros/melodic/share/control_toolbox',
        '/opt/ros/melodic/share/costmap_2d',
        '/opt/ros/melodic/share/cpp_common',
        '/opt/ros/melodic/share/cv_bridge',
        '/opt/ros/melodic/share/depth_image_proc',
        '/opt/ros/melodic/share/diagnostic_msgs',
        '/opt/ros/melodic/share/diagnostic_updater',
        '/opt/ros/melodic/share/dynamic_reconfigure',
        '/opt/ros/melodic/share/eigen_conversions',
        '/opt/ros/melodic/share/eigen_stl_containers',
        '/opt/ros/melodic/share/eigenpy',
        '/ros_ws/src/fetch_ros/fetch_depth_layer',
        '/ros_ws/src/fetch_ros/fetch_description',
        '/ros_ws/src/fetch_gazebo/fetch_gazebo',
        '/ros_ws/src/fetch_gazebo/fetch_gazebo_demo',
        '/ros_ws/src/fetch_ros/fetch_ikfast_plugin',
        '/ros_ws/src/fetch_ros/fetch_maps',
        '/ros_ws/src/fetch_ros/fetch_moveit_config',
        '/ros_ws/src/fetch_ros/fetch_navigation',
        '/opt/ros/melodic/share/gazebo_dev',
        '/opt/ros/melodic/share/gazebo_msgs',
        '/opt/ros/melodic/share/gazebo_plugins',
        '/opt/ros/melodic/share/gazebo_ros',
        '/opt/ros/melodic/share/gencpp',
        '/opt/ros/melodic/share/geneus',
        '/opt/ros/melodic/share/genlisp',
        '/opt/ros/melodic/share/genmsg',
        '/opt/ros/melodic/share/gennodejs',
        '/opt/ros/melodic/share/genpy',
        '/opt/ros/melodic/share/geometric_shapes',
        '/opt/ros/melodic/share/geometry_msgs',
        '/opt/ros/melodic/share/grasping_msgs',
        '/opt/ros/melodic/share/image_geometry',
        '/opt/ros/melodic/share/image_proc',
        '/opt/ros/melodic/share/image_transport',
        '/opt/ros/melodic/share/interactive_markers',
        '/opt/ros/melodic/share/joint_state_publisher',
        '/opt/ros/melodic/share/kdl_conversions',
        '/opt/ros/melodic/share/kdl_parser',
        '/opt/ros/melodic/share/laser_geometry',
        '/opt/ros/melodic/share/map_msgs',
        '/opt/ros/melodic/share/map_server',
        '/opt/ros/melodic/share/media_export',
        '/opt/ros/melodic/share/message_filters',
        '/opt/ros/melodic/share/message_generation',
        '/opt/ros/melodic/share/message_runtime',
        '/opt/ros/melodic/share/mk',
        '/opt/ros/melodic/share/move_base',
        '/opt/ros/melodic/share/move_base_msgs',
        '/opt/ros/melodic/share/moveit_commander',
        '/opt/ros/melodic/share/moveit_core',
        '/opt/ros/melodic/share/moveit_fake_controller_manager',
        '/opt/ros/melodic/share/moveit_kinematics',
        '/opt/ros/melodic/share/moveit_msgs',
        '/opt/ros/melodic/share/moveit_planners_ompl',
        '/opt/ros/melodic/share/moveit_python',
        '/opt/ros/melodic/share/moveit_ros_manipulation',
        '/opt/ros/melodic/share/moveit_ros_move_group',
        '/opt/ros/melodic/share/moveit_ros_perception',
        '/opt/ros/melodic/share/moveit_ros_planning',
        '/opt/ros/melodic/share/moveit_ros_planning_interface',
        '/opt/ros/melodic/share/moveit_ros_robot_interaction',
        '/opt/ros/melodic/share/moveit_ros_visualization',
        '/opt/ros/melodic/share/moveit_ros_warehouse',
        '/opt/ros/melodic/share/moveit_simple_controller_manager',
        '/opt/ros/melodic/share/nav_core',
        '/opt/ros/melodic/share/nav_msgs',
        '/opt/ros/melodic/share/navfn',
        '/opt/ros/melodic/share/nodelet',
        '/opt/ros/melodic/share/nodelet_topic_tools',
        '/opt/ros/melodic/share/object_recognition_msgs',
        '/opt/ros/melodic/share/octomap',
        '/opt/ros/melodic/share/octomap_msgs',
        '/opt/ros/melodic/share/ompl',
        '/opt/ros/melodic/share/open_karto',
        '/opt/ros/melodic/share/orocos_kdl',
        '/opt/ros/melodic/share/pcl_conversions',
        '/opt/ros/melodic/share/pcl_msgs',
        '/opt/ros/melodic/share/pcl_ros',
        '/opt/ros/melodic/share/pluginlib',
        '/opt/ros/melodic/share/polled_camera',
        '/opt/ros/melodic/share/python_orocos_kdl',
        '/opt/ros/melodic/share/python_qt_binding',
        '/opt/ros/melodic/share/random_numbers',
        '/opt/ros/melodic/share/realtime_tools',
        '/opt/ros/melodic/share/resource_retriever',
        '/opt/ros/melodic/share/rgbd_launch',
        '/opt/ros/melodic/share/robot_controllers',
        '/opt/ros/melodic/share/robot_controllers_interface',
        '/opt/ros/melodic/share/robot_controllers_msgs',
        '/opt/ros/melodic/share/robot_state_publisher',
        '/opt/ros/melodic/share/ros_environment',
        '/opt/ros/melodic/share/rosbag',
        '/opt/ros/melodic/share/rosbag_migration_rule',
        '/opt/ros/melodic/share/rosbag_storage',
        '/opt/ros/melodic/share/rosbash',
        '/opt/ros/melodic/share/rosboost_cfg',
        '/opt/ros/melodic/share/rosbuild',
        '/opt/ros/melodic/share/rosclean',
        '/opt/ros/melodic/share/rosconsole',
        '/opt/ros/melodic/share/rosconsole_bridge',
        '/opt/ros/melodic/share/roscpp',
        '/opt/ros/melodic/share/roscpp_serialization',
        '/opt/ros/melodic/share/roscpp_traits',
        '/opt/ros/melodic/share/roscreate',
        '/opt/ros/melodic/share/rosgraph',
        '/opt/ros/melodic/share/rosgraph_msgs',
        '/opt/ros/melodic/share/roslang',
        '/opt/ros/melodic/share/roslaunch',
        '/opt/ros/melodic/share/roslib',
        '/opt/ros/melodic/share/roslisp',
        '/opt/ros/melodic/share/roslz4',
        '/opt/ros/melodic/share/rosmake',
        '/opt/ros/melodic/share/rosmaster',
        '/opt/ros/melodic/share/rosmsg',
        '/opt/ros/melodic/share/rosnode',
        '/opt/ros/melodic/share/rosout',
        '/opt/ros/melodic/share/rospack',
        '/opt/ros/melodic/share/rosparam',
        '/opt/ros/melodic/share/rospy',
        '/opt/ros/melodic/share/rosservice',
        '/opt/ros/melodic/share/rostest',
        '/opt/ros/melodic/share/rostime',
        '/opt/ros/melodic/share/rostopic',
        '/opt/ros/melodic/share/rosunit',
        '/opt/ros/melodic/share/roswtf',
        '/opt/ros/melodic/share/rotate_recovery',
        '/opt/ros/melodic/share/rviz',
        '/opt/ros/melodic/share/sensor_msgs',
        '/opt/ros/melodic/share/shape_msgs',
        '/opt/ros/melodic/share/simple_grasping',
        '/opt/ros/melodic/share/slam_karto',
        '/opt/ros/melodic/share/smclib',
        '/opt/ros/melodic/share/sparse_bundle_adjustment',
        '/opt/ros/melodic/share/srdfdom',
        '/opt/ros/melodic/share/std_msgs',
        '/opt/ros/melodic/share/std_srvs',
        '/opt/ros/melodic/share/stereo_msgs',
        '/opt/ros/melodic/share/teleop_twist_keyboard',
        '/opt/ros/melodic/share/tf',
        '/opt/ros/melodic/share/tf2',
        '/opt/ros/melodic/share/tf2_eigen',
        '/opt/ros/melodic/share/tf2_geometry_msgs',
        '/opt/ros/melodic/share/tf2_kdl',
        '/opt/ros/melodic/share/tf2_msgs',
        '/opt/ros/melodic/share/tf2_py',
        '/opt/ros/melodic/share/tf2_ros',
        '/opt/ros/melodic/share/tf_conversions',
        '/opt/ros/melodic/share/topic_tools',
        '/opt/ros/melodic/share/trajectory_msgs',
        '/opt/ros/melodic/share/urdf',
        '/opt/ros/melodic/share/urdfdom_py',
        '/opt/ros/melodic/share/visualization_msgs',
        '/opt/ros/melodic/share/voxel_grid',
        '/opt/ros/melodic/share/warehouse_ros',
        '/opt/ros/melodic/share/xacro',
        '/opt/ros/melodic/share/xmlrpcpp'
    }
    actual = set(PackageDatabase.paths(sut.shell, sut.files))
    assert actual == expected


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_database_from_paths(filesystem):
    paths = [
        '/opt/ros/melodic/share/angles',
        '/opt/ros/melodic/share/tf2',
        '/opt/ros/melodic/share/tf2_msgs',
        '/opt/ros/melodic/share/tf2_py',
        '/opt/ros/melodic/share/tf2_ros'
    ]
    db = PackageDatabase.from_paths(filesystem, paths)
    assert len(db) == len(paths)
    assert set(db) == {'angles', 'tf2', 'tf2_msgs', 'tf2_py', 'tf2_ros'}



# -*- coding: utf-8 -*-
__all__ = ('NodeConfig',)

from typing import Collection, Optional, Sequence, Tuple
import xml.etree.ElementTree as ET

import attr

from ....name import namespace_join


@attr.s(frozen=True, slots=True, auto_attribs=True)
class NodeConfig:
    namespace: str
    name: str
    typ: str
    package: str
    remappings: Collection[Tuple[str, str]] = attr.ib(default=tuple())
    filename: Optional[str] = attr.ib(default=None)
    output: Optional[str] = attr.ib(default=None)
    required: bool = attr.ib(default=False)
    respawn: bool = attr.ib(default=False)
    respawn_delay: float = attr.ib(default=0.0)
    env_args: Sequence[Tuple[str, str]] = attr.ib(default=tuple())
    cwd: Optional[str] = attr.ib(default=None)
    args: Optional[str] = attr.ib(default=None)
    launch_prefix: Optional[str] = attr.ib(default=None)

    @property
    def full_name(self) -> str:
        return namespace_join(self.namespace, self.name)

    def with_remappings(self,
                        remappings: Collection[Tuple[str, str]],
                        *,
                        overwrite: bool = False
                        ) -> 'NodeConfig':
        """Applies a set of remappings to this configuration.

        Parameters
        ----------
        remappings: Collection[Tuple[str, str]]
            A collection of name remappings, each given as a tuple of the
            form :code:`(from, to)`.
        overwrite: bool
            If :code:`True`, any existing remappings will not appear in the
            returned variant. Otherwise, the returned variant will contain
            both the supplied and existing remappings.

        Returns
        -------
        NodeConfig
            A variant of this configuration with the given remappings.
        """
        if not overwrite:
            remappings = tuple(self.remappings) + tuple(remappings)
        return attr.evolve(self, remappings=remappings)

    def to_xml_element(self) -> ET.Element:
        element = ET.Element('node')
        element.attrib['pkg'] = self.package
        element.attrib['type'] = self.typ
        element.attrib['name'] = self.name
        element.attrib['ns'] = self.namespace
        element.attrib['respawn'] = str(self.respawn)
        element.attrib['respawn_delay'] = str(self.respawn_delay)
        element.attrib['required'] = str(self.required)
        if self.launch_prefix:
            element.attrib['launch-prefix'] = self.launch_prefix
        if self.args:
            element.attrib['args'] = self.args
        if self.cwd:
            element.attrib['cwd'] = self.cwd
        if self.output:
            element.attrib['output'] = self.output
        for remap_from, remap_to in self.remappings:
            attrib = {'from': remap_from, 'to': remap_to}
            ET.SubElement(element, 'remap', attrib=attrib)
        return element



# -*- coding: utf-8 -*-
__all__ = ('ROSCore',)

from typing import Dict, Sequence, Optional, Tuple
import os
import xmlrpc.client
import time

from loguru import logger
import dockerblade

from ..description import SystemDescription
from ..exceptions import ROSWireException
from .bag import BagRecorder, BagPlayer
from .node import NodeManager
from .parameters import ParameterServer
from .roslaunch import ROSLaunchManager
from .service import ServiceManager
from .state import SystemState, SystemStateProbe


class ROSCore:
    """Provides access to a remote ROS master via XML-RPC.

    Attributes
    ----------
    uri: str
        The URI of the ROS Master.
    connection: xmlrpc.client.ServerProxy
        The XML-RPC connection to the ROS Master.
    nodes: NodeManager
        Provides access to the nodes running on this ROS Master.
    state: SystemState
        The instantaneous state of the ROS Master.
    roslaunch: ROSLaunchManager
        Provides access to roslaunch-related functionality.
    services: ServiceManager
        Provides access to the services advertised on this ROS Master.
    parameters: ParameterServer
        Provides access to the parameter server for this ROS Master.
    topic_to_type: Dict[str, str]
        A mapping from topic names to the names of their message types.
    """
    def __init__(self,
                 description: SystemDescription,
                 shell: dockerblade.Shell,
                 files: dockerblade.FileSystem,
                 ws_host: str,
                 ip_address: str,
                 port: int = 11311
                 ) -> None:
        self.__description = description
        self.__shell = shell
        self.__files = files
        self.__ws_host = ws_host
        self.__caller_id = '/roswire'
        self.__port = port
        self.__ip_address = ip_address
        self.__uri = f"http://{ip_address}:{port}"
        logger.debug(f"connecting to ROS Master: {self.__uri}")
        self.__connection = xmlrpc.client.ServerProxy(self.__uri)
        time.sleep(5)  # FIXME #1
        self.__parameters = ParameterServer(self.__connection)
        self.__nodes: NodeManager = \
            NodeManager(self.__ip_address,
                        self.__connection,
                        self.__shell)
        self.__services: ServiceManager = \
            ServiceManager(self.__description,
                           self.__ip_address,
                           self.__connection,
                           self.__shell)
        self.__state_probe: SystemStateProbe = \
            SystemStateProbe.via_xmlrpc_connection(self.__connection)
        self.roslaunch: ROSLaunchManager = \
            ROSLaunchManager(self.__shell, self.__files)

    @property
    def nodes(self) -> NodeManager:
        return self.__nodes

    @property
    def services(self) -> ServiceManager:
        return self.__services

    @property
    def parameters(self) -> ParameterServer:
        return self.__parameters

    @property
    def connection(self) -> xmlrpc.client.ServerProxy:
        return self.__connection

    @property
    def state(self) -> SystemState:
        return self.__state_probe()

    @property
    def topic_to_type(self) -> Dict[str, str]:
        code: int
        msg: str
        result: Sequence[Tuple[str, str]]
        code, msg, result = \
            self.connection.getTopicTypes(self.__caller_id)  # type: ignore
        if code != 1:
            raise ROSWireException("bad API call!")
        return {name: typ for (name, typ) in result}

    def record(self,
               fn: str,
               exclude_topics: Optional[str] = None,
               restrict_to_topics: Optional[str] = None
               ) -> BagRecorder:
        """Provides an interface to rosbag for recording ROS topics to disk.

        Note
        ----
        This method records bag files to the host machine, and not to the
        container where the ROS instance is running.

        Parameters
        ----------
        fn: str
            The name of the file, on the host machine, to which the bag should
            be recorded
        exclude_topics: str, optional
            An optional regular expression specifying the topics that should
            be excluded from the bag.
        restrict_to_topics: str, optional
            An optional regular expression specifying the topics to which
            recording should be restricted.

        Returns
        -------
        BagRecorder
            An interface for dynamically interacting with the bag recorder.
        """
        return BagRecorder(fn,
                           self.__ws_host,
                           self.__shell,
                           self.__nodes,
                           exclude_topics=exclude_topics,
                           restrict_to_topics=restrict_to_topics)

    def playback(self,
                 fn: str,
                 *,
                 file_on_host: bool = True
                 ) -> BagPlayer:
        """Provides an interface to rosbag for replaying bag files from disk.

        Parameters
        ----------
        fn: str
            The bag file that should be replayed.
        file_on_host: bool
            If :code:`True`, as by default, :code:`fn` will be considered to
            be a file on the host machine. If :code:`False`, :code:`fn` will
            be considered to be a file inside the container.

        Returns
        -------
        BagPlayer
            An interface for dynamically controlling the bag player.
        """
        fn_ctr: str
        delete_file_after_use: bool = False
        if file_on_host:
            if fn.startswith(self.__ws_host):
                fn_ctr = os.path.join('/.roswire', fn[len(self.__ws_host):])
            else:
                delete_file_after_use = True
                fn_ctr = self.__files.mktemp(suffix='.bag')
                logger.debug(f"copying bag from host [{fn}] "
                             f"to container [{fn_ctr}]")
                self.__files.copy_from_host(fn, fn_ctr)
        else:
            fn_ctr = fn
        logger.debug(f"playing back bag file: {fn_ctr}")
        return BagPlayer(fn_ctr,
                         self.__shell,
                         self.__files,
                         delete_file_after_use=delete_file_after_use)



# -*- coding: utf-8 -*-
"""
This module provides access to the ROSWire session.
"""
__all__ = ('ROSWire',)

from typing import Dict, Iterator, Mapping, Optional, Sequence
import os
import contextlib

from loguru import logger

from .exceptions import ROSWireException
from .description import SystemDescription, SystemDescriptionManager
from .system import System
from .proxy import ContainerManager


class ROSWire:
    """
    Provides an interface for building, analysing, and interacting with
    containerised ROS applications.

    Attributes
    ----------
    containers: ContainerManager
        A manager for building and connecting to Docker containers.
    descriptions: SystemDescriptionManager
        A manager for building, loading, and storing static descriptions of
        ROS applications.
    workspace: str
        The absolute path of the workspace directory for this session. The
        workspace is used to store cache data and to store shared temporary
        directories between the host machine and containerised ROS
        applications.
    """
    def __init__(self,
                 dir_workspace: Optional[str] = None
                 ) -> None:
        if not dir_workspace:
            logger.debug("no workspace specified: using default workspace.")
            dir_home = os.path.expanduser("~")
            dir_workspace = os.path.join(dir_home, ".roswire")
            logger.debug(f"default workspace: {dir_workspace}")
            if not os.path.exists(dir_workspace):
                logger.debug("initialising default workspace")
                os.mkdir(dir_workspace)
        else:
            logger.debug(f"using specified workspace: {dir_workspace}")
            if not os.path.exists(dir_workspace):
                m = f"workspace not found: {dir_workspace}"
                raise ROSWireException(m)

        self.__dir_workspace = os.path.abspath(dir_workspace)
        self.__containers = ContainerManager(self.__dir_workspace)
        dir_descriptions = os.path.join(dir_workspace, 'descriptions')
        self.__descriptions = SystemDescriptionManager(self.__containers,
                                                       dir_descriptions)

    @property
    def workspace(self) -> str:
        return self.__dir_workspace

    @property
    def containers(self) -> ContainerManager:
        return self.__containers

    @property
    def descriptions(self) -> SystemDescriptionManager:
        return self.__descriptions

    @contextlib.contextmanager
    def launch(self,
               image: str,
               sources: Sequence[str],
               description: Optional[SystemDescription] = None,
               *,
               ports: Optional[Dict[int, int]] = None,
               environment: Optional[Mapping[str, str]] = None
               ) -> Iterator[System]:
        """Launches a ROS application using a provided Docker image.

        Parameters
        ----------
        image: str
            the name of the Docker image.
        sources: Sequence[str]
            The sequence of setup files that should be used to load the ROS
            workspace.
        description: Optional[SystemDescription]
            an optional static description of the ROS application.
            If no description is provided, ROSWire will attempt to load one
            from the cache or else build one.
        ports: Dict[int, int], optional
            an optional dictionary specifying port mappings between the host
            and container, where keys represent container ports and values
            represent host ports.
        environment: Mapping[str, str], optional
            an optional set of additional environment variables, indexed by
            name, that should be used by the system.
        """
        if not description:
            description = self.descriptions.load_or_build(image, sources)
        with self.containers.launch(image,
                                    ports=ports,
                                    sources=sources,
                                    environment=environment) as container:
            container = container
            yield System(container, description)



# -*- coding: utf-8 -*-
import pytest

import roswire


@pytest.mark.skip(reason='this takes 10-15 minutes')
def test_description():
    NAME_IMAGE = 'hello-world'
    rsw = roswire.ROSWire()
    desc = rsw.descriptions.build(NAME_IMAGE, save=False)
    assert set(desc.packages) == {
        'nodelet_core',
        'ros_core',
        'ros_comm',
        'ros_base',
        'ros_tutorials',
        'common_tutorials',
        'roscpp_core',
        'bond_core',
        'ros',
        'common_msgs',
        'actionlib',
        'actionlib_msgs',
        'actionlib_tutorials',
        'angles',
        'bond',
        'bondcpp',
        'bondpy',
        'catkin',
        'class_loader',
        'cmake_modules',
        'cpp_common',
        'diagnostic_msgs',
        'dynamic_reconfigure',
        'gencpp',
        'genlisp',
        'genmsg',
        'genpy',
        'geometry_msgs',
        'message_filters',
        'message_generation',
        'message_runtime',
        'mk',
        'nav_msgs',
        'nodelet',
        'nodelet_topic_tools',
        'nodelet_tutorial_math',
        'pluginlib',
        'pluginlib_tutorials',
        'rosbag',
        'rosbag_migration_rule',
        'rosbag_storage',
        'rosbash',
        'rosboost_cfg',
        'rosbuild',
        'rosclean',
        'rosconsole',
        'rosconsole_bridge',
        'roscpp',
        'roscpp_serialization',
        'roscpp_traits',
        'roscpp_tutorials',
        'roscreate',
        'rosgraph',
        'rosgraph_msgs',
        'roslang',
        'roslaunch',
        'roslib',
        'roslisp',
        'roslz4',
        'rosmake',
        'rosmaster',
        'rosmsg',
        'rosnode',
        'rosout',
        'rospack',
        'rosparam',
        'rospy',
        'rospy_tutorials',
        'rosservice',
        'rostest',
        'rostime',
        'rostopic',
        'rosunit',
        'roswtf',
        'sensor_msgs',
        'shape_msgs',
        'smclib',
        'std_msgs',
        'std_srvs',
        'stereo_msgs',
        'topic_tools',
        'trajectory_msgs',
        'turtle_actionlib',
        'turtlesim',
        'visualization_msgs',
        'xmlrpcpp'}



# -*- coding: utf-8 -*-
"""
This module provides code for encoding and serialising Python data structures
into their corresponding ROS binary representations.
"""
from typing import Optional, Callable, Any, TypeVar, BinaryIO, Sequence, Dict
import functools
import struct

from .base import Time, Duration
from .decode import get_pattern

T = TypeVar('T')


def ignore(val: Any) -> None:
    """Used to prevent returning values in lambdas."""
    return None


def simple_encoder(typ: str) -> Callable[[Any], bytes]:
    """Returns an encoder for a specified simple type."""
    pattern = '<' + get_pattern(typ)
    return functools.partial(struct.pack, pattern)


def sized_encoder(encoder_content: Callable[[T], bytes],
                  get_size: Callable[[T], int]
                  ) -> Callable[[T], bytes]:
    def encoder(val: T) -> bytes:
        bin_size = encode_uint32(get_size(val))
        bin_content = encoder_content(val)
        return bin_size + bin_content

    return encoder


def writer(encoder: Callable[[Any], bytes]
           ) -> Callable[[Any, BinaryIO], None]:
    return lambda v, b: ignore(b.write(encoder(v)))


def simple_writer(typ: str) -> Callable[[Any, BinaryIO], None]:
    """Returns a writer for a specified simple type."""
    return writer(simple_encoder(typ))


encode_int8 = simple_encoder('int8')
encode_uint8 = simple_encoder('uint8')
encode_int16 = simple_encoder('int16')
encode_uint16 = simple_encoder('uint16')
encode_int32 = simple_encoder('int32')
encode_uint32 = simple_encoder('uint32')
encode_int64 = simple_encoder('int64')
encode_uint64 = simple_encoder('uint64')
encode_float32 = simple_encoder('float32')
encode_float64 = simple_encoder('float64')
encode_char = simple_encoder('char')
encode_byte = simple_encoder('byte')
encode_bool = simple_encoder('bool')


def encode_time(time: Time) -> bytes:
    return encode_uint32(time.secs) + encode_uint32(time.nsecs)


def encode_duration(duration: Duration) -> bytes:
    return encode_uint32(duration.secs) + encode_uint32(duration.nsecs)


write_int8 = simple_writer('int8')
write_uint8 = simple_writer('uint8')
write_int16 = simple_writer('int16')
write_uint16 = simple_writer('uint16')
write_int32 = simple_writer('int32')
write_uint32 = simple_writer('uint32')
write_int64 = simple_writer('int64')
write_uint64 = simple_writer('uint64')
write_float32 = simple_writer('float32')
write_float64 = simple_writer('float64')
write_char = simple_writer('char')
write_byte = simple_writer('byte')
write_bool = simple_writer('bool')
write_time = writer(encode_time)
write_duration = writer(encode_duration)


def write_encoded_header(contents: Dict[str, bytes], out: BinaryIO) -> None:
    # write a length of zero for now and correct that length once the
    # fields have been written
    pos_size = out.tell()
    write_uint32(0, out)
    pos_content = out.tell()

    for name, bin_value in contents.items():
        bin_name = f'{name}='.encode('utf-8')
        size_field = len(bin_name) + len(bin_value)
        write_uint32(size_field, out)
        out.write(bin_name)
        out.write(bin_value)

    # correct the length
    pos_end = out.tell()
    size_total = pos_end - pos_content
    out.seek(pos_size)
    write_uint32(size_total, out)
    out.seek(pos_end)


def string_writer(length: Optional[int] = None
                  ) -> Callable[[str, BinaryIO], None]:
    """Returns a writer for (possibly fixed-length) strings."""
    encoder: Callable[[str], bytes]
    encode_content: Callable[[str], bytes] = str.encode
    if length is None:
        encoder = sized_encoder(encode_content, str.__len__)
    else:
        encoder = encode_content
    return writer(encoder)


def simple_array_writer(typ: str,
                        length: Optional[int] = None
                        ) -> Callable[[Sequence[Any], BinaryIO], None]:
    """Returns a writer for a simple array."""
    base_pattern = get_pattern(typ)

    if length is not None:
        pattern = f'<{length}{base_pattern}'
        return lambda arr, b: ignore(struct.pack(pattern, *arr))

    def var_writer(arr: Sequence[Any], b: BinaryIO) -> None:
        length = len(arr)
        write_uint32(length, b)
        pattern = f'<{length}{base_pattern}'
        b.write(struct.pack(pattern, *arr))

    return var_writer


def complex_array_writer(entry_writer: Callable[[Any, BinaryIO], None],
                         length: Optional[int] = None
                         ) -> Callable[[Any, BinaryIO], None]:
    """Returns a writer for a complex array."""
    def write_content(arr: Sequence[Any], b: BinaryIO) -> None:
        for v in arr:
            entry_writer(v, b)

    if length is not None:
        return write_content

    def var_writer(arr: Sequence[Any], b: BinaryIO) -> None:
        length = len(arr)
        write_uint32(length, b)
        write_content(arr, b)

    return var_writer



# -*- coding: utf-8 -*-
import pytest

from typing import List
import os

from roswire.definitions import (Constant, Field, MsgFormat, SrvFormat,
                                 ActionFormat, Time, Package, PackageDatabase,
                                 FormatDatabase)
import dockerblade


def test_msg_from_string():
    s = """
#Standard metadata for higher-level flow data types
#sequence ID: consecutively increasing ID
uint32 seq
#Two-integer timestamp that is expressed as:
# * stamp.secs: seconds (stamp_secs) since epoch
# * stamp.nsecs: nanoseconds since stamp_secs
# time-handling sugar is provided by the client library
time stamp
#Frame this data is associated with
string frame_id

int32 X=123
int32 Y=-123
string FOO=foo
string EXAMPLE="#comments" are ignored, and leading and trailing whitespace removed
    """
    fmt = MsgFormat.from_string("PkgName", "MessageName", s)
    assert fmt.name == "MessageName"
    assert fmt.package == "PkgName"

    assert len(fmt.constants) == 4
    assert Constant('int32', 'X', '123') in fmt.constants
    assert Constant('int32', 'Y', '-123') in fmt.constants
    assert Constant('string', 'FOO', 'foo') in fmt.constants
    assert Constant('string', 'EXAMPLE', '"#comments" are ignored, and leading and trailing whitespace removed') in fmt.constants  # noqa: pycodestyle

    assert len(fmt.fields) == 3
    assert Field('uint32', 'seq') in fmt.fields
    assert Field('time', 'stamp') in fmt.fields
    assert Field('string', 'frame_id') in fmt.fields


def test_srv_from_string():
    s = """
#request constants
int8 FOO=1
int8 BAR=2
#request fields
int8 foobar
another_pkg/AnotherMessage msg
---
#response constants
uint32 SECRET=123456
#response fields
another_pkg/YetAnotherMessage val
CustomMessageDefinedInThisPackage value
uint32 an_integer
    """

    fmt = SrvFormat.from_string("PkgName", "MessageName", s)
    assert fmt.name == "MessageName"
    assert fmt.package == "PkgName"

    # check request
    req = fmt.request
    assert req is not None

    assert len(req.constants) == 2
    assert Constant('int8', 'FOO', '1') in req.constants
    assert Constant('int8', 'BAR', '2') in req.constants

    assert len(req.fields) == 2
    assert Field('int8', 'foobar') in req.fields
    assert Field('another_pkg/AnotherMessage', 'msg') in req.fields

    # check response
    res = fmt.response
    assert res is not None

    assert len(res.constants) == 1
    assert Constant('uint32', 'SECRET', '123456') in res.constants

    assert len(res.fields) == 3
    assert Field('another_pkg/YetAnotherMessage', 'val') in res.fields
    assert Field('PkgName/CustomMessageDefinedInThisPackage', 'value') in res.fields
    assert Field('uint32', 'an_integer') in res.fields

    # bug #269
    s = "map_msgs/ProjectedMapInfo[] projected_maps_info"
    fmt = SrvFormat.from_string("map_msgs", "ProjectedMapsInfo", s)
    assert fmt.request is not None
    assert fmt.response is None


def test_action_from_string():
    s = """
# Define the goal
uint32 dishwasher_id  # Specify which dishwasher we want to use
---
# Define the result
uint32 total_dishes_cleaned
---
# Define a feedback message
float32 percent_complete
"""

    fmt = ActionFormat.from_string("PkgName", "MessageName", s)
    assert fmt.name == "MessageName"
    assert fmt.package == "PkgName"

    # check goal
    goal = fmt.goal
    assert goal is not None
    assert not goal.constants
    assert len(goal.fields) == 1
    assert Field('uint32', 'dishwasher_id') in goal.fields

    # check result
    res = fmt.result
    assert res is not None
    assert not res.constants
    assert len(res.fields) == 1
    assert Field('uint32', 'total_dishes_cleaned') in res.fields

    # check feedback
    feedback = fmt.feedback
    assert feedback is not None
    assert not feedback.constants
    assert len(feedback.fields) == 1
    assert Field('float32', 'percent_complete') in feedback.fields


def test_empty_action_from_string():
    """see #332"""
    s = "---\n---\n"
    fmt = ActionFormat.from_string("PkgName", "MessageName", s)
    assert fmt.goal is not None
    assert not fmt.goal.fields

    assert fmt.result is not None
    assert not fmt.result.fields

    assert fmt.feedback is not None
    assert not fmt.feedback.fields


def test_field_to_and_from_dict():
    d = {'type': 'uint32', 'name': 'foo'}
    f = Field('uint32', 'foo')
    assert Field.from_dict(d) == Field('uint32', 'foo')
    assert Field.from_dict(f.to_dict()) == f


def test_constant_to_and_from_dict():
    d = {'type': 'uint32', 'name': 'foo', 'value': '100'}
    c = Constant('uint32', 'foo', '100')
    assert Constant.from_dict(d) == c
    assert Constant.from_dict(c.to_dict()) == c


def test_msg_format_to_and_from_dict():
    definition = 'geometry_msgs/TransformStamped[] transforms\n'
    d = {'package': 'tf',
         'name': 'tfMessage',
         'definition': definition,
         'fields': [
             {'type': 'geometry_msgs/TransformStamped[]',
              'name': 'transforms'}]}
    f = MsgFormat(package='tf',
                  name='tfMessage',
                  definition=definition,
                  constants=[],
                  fields=[Field('geometry_msgs/TransformStamped[]', 'transforms')])
    assert MsgFormat.from_dict(d) == f
    assert MsgFormat.from_dict(f.to_dict()) == f


def test_srv_format_to_and_from_dict():
    pkg = 'nav_msgs'
    name = 'SetMap'
    name_request = 'SetMapRequest'
    name_response = 'SetMapResponse'
    # it doesn't matter what these are set to
    definition_service = 'FOO'
    definition_request = 'BAR'
    definition_response = 'COOL'
    d = {'package': pkg,
         'name': name,
         'definition': definition_service,
         'request': {
            'definition': definition_request,
            'fields': [
                {'type': 'nav_msgs/OccupancyGrid',
                 'name': 'map'},
                {'type': 'geometry_msgs/PoseWithCovarianceStamped',
                 'name': 'initial_pose'}]
         },
         'response': {
            'definition': definition_response,
            'fields': [{'type': 'bool', 'name': 'success'}]
         }}
    f = SrvFormat(
            package=pkg,
            name=name,
            definition=definition_service,
            request=MsgFormat(
                package=pkg,
                definition=definition_request,
                name=name_request,
                constants=[],
                fields=[Field('nav_msgs/OccupancyGrid', 'map'),
                        Field('geometry_msgs/PoseWithCovarianceStamped',
                              'initial_pose')]),
            response=MsgFormat(
                package=pkg,
                definition=definition_response,
                name=name_response,
                constants=[],
                fields=[Field('bool', 'success')]))
    assert SrvFormat.from_dict(d) == f
    assert SrvFormat.from_dict(f.to_dict()) == f


def test_action_format_to_and_from_dict():
    pkg = 'actionlib'
    name = 'TwoInts'
    name_goal = 'TwoIntsGoal'
    name_result = 'TwoIntsResult'
    # it doesn't matter what these are set to
    definition_action = 'FOO'
    definition_goal = 'BAR'
    definition_result = 'COOL'
    d = {'package': pkg,
         'name': name,
         'definition': definition_action,
         'goal': {
            'definition': definition_goal,
            'fields': [{'type': 'int64', 'name': 'a'},
                       {'type': 'int64', 'name': 'b'}]
         },
         'result': {
            'definition': definition_result,
            'fields': [{'type': 'int64', 'name': 'sum'}]
         }}
    f = ActionFormat(
            package=pkg,
            name=name,
            definition=definition_action,
            goal=MsgFormat(
                definition=definition_goal,
                package=pkg,
                name=name_goal,
                constants=[],
                fields=[Field('int64', 'a'),
                        Field('int64', 'b')]),
            feedback=None,
            result=MsgFormat(
                package=pkg,
                definition=definition_result,
                name=name_result,
                constants=[],
                fields=[Field('int64', 'sum')]))
    assert ActionFormat.from_dict(d) == f
    assert ActionFormat.from_dict(f.to_dict()) == f


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_action_from_file(filesystem):
    # read .action file
    pkg = 'tf2_msgs'
    pkg_dir = '/opt/ros/melodic/share/tf2_msgs'
    fn = os.path.join(pkg_dir, 'action/LookupTransform.action')
    fmt = ActionFormat.from_file(pkg, fn, filesystem)
    assert fmt.package == pkg
    assert fmt.name == 'LookupTransform'
    assert fmt.fullname == 'tf2_msgs/LookupTransform'

    goal: MsgFormat = fmt.goal
    assert not goal.constants
    assert len(goal.fields) == 7
    assert Field('string', 'target_frame') in goal.fields
    assert Field('string', 'source_frame') in goal.fields
    assert Field('time', 'source_time') in goal.fields
    assert Field('duration', 'timeout') in goal.fields
    assert Field('time', 'target_time') in goal.fields
    assert Field('string', 'fixed_frame') in goal.fields
    assert Field('bool', 'advanced') in goal.fields

    assert fmt.result
    res: MsgFormat = fmt.result
    assert not res.constants
    assert len(res.fields) == 2
    assert Field('geometry_msgs/TransformStamped', 'transform') in res.fields
    assert Field('tf2_msgs/TF2Error', 'error') in res.fields

    # attempt to read .msg file
    fn = os.path.join(pkg_dir, 'msg/TFMessage.msg')
    with pytest.raises(AssertionError):
        ActionFormat.from_file(pkg, fn, filesystem)

    # attempt to read non-existent file
    fn = os.path.join(pkg_dir, 'action/Spooky.action')
    with pytest.raises(dockerblade.exceptions.ContainerFileNotFound):
        ActionFormat.from_file(pkg, fn, filesystem)


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_srv_from_file(filesystem):
    # read .srv file
    pkg = 'nav_msgs'
    pkg_dir = '/opt/ros/melodic/share/nav_msgs'
    fn = os.path.join(pkg_dir, 'srv/SetMap.srv')
    fmt = SrvFormat.from_file(pkg, fn, filesystem)
    assert fmt.package == pkg
    assert fmt.name == 'SetMap'
    assert fmt.fullname == 'nav_msgs/SetMap'

    req: MsgFormat = fmt.request
    assert not req.constants
    assert len(req.fields) == 2
    assert Field('nav_msgs/OccupancyGrid', 'map') in req.fields
    assert Field('geometry_msgs/PoseWithCovarianceStamped', 'initial_pose') in req.fields

    assert fmt.response
    res: MsgFormat = fmt.response
    assert not res.constants
    assert len(res.fields) == 1
    assert Field('bool', 'success') in res.fields

    # attempt to read .action file
    fn = '/opt/ros/melodic/share/tf2_msgs/action/LookupTransform.action'
    with pytest.raises(AssertionError):
        SrvFormat.from_file(pkg, fn, filesystem)

    # attempt to read non-existent file
    fn = os.path.join(pkg_dir, 'srv/Spooky.srv')
    with pytest.raises(dockerblade.exceptions.ContainerFileNotFound):
        SrvFormat.from_file(pkg, fn, filesystem)


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_msg_from_file(filesystem):
    # read .msg file
    pkg = 'tf2_msgs'
    pkg_dir = '/opt/ros/melodic/share/tf2_msgs/'
    fn = os.path.join(pkg_dir, 'msg/TFMessage.msg')
    fmt = MsgFormat.from_file(pkg, fn, filesystem)
    assert fmt.package == pkg
    assert fmt.name == 'TFMessage'
    assert fmt.fullname == 'tf2_msgs/TFMessage'
    assert not fmt.constants
    assert len(fmt.fields) == 1
    assert Field('geometry_msgs/TransformStamped[]', 'transforms') in fmt.fields

    # attempt to read .action file
    fn = os.path.join(pkg_dir, 'action/LookupTransform.action')
    with pytest.raises(AssertionError):
        SrvFormat.from_file(pkg, fn, filesystem)

    # attempt to read non-existent file
    fn = os.path.join(pkg_dir, 'msg/Spooky.msg')
    with pytest.raises(dockerblade.exceptions.ContainerFileNotFound):
        MsgFormat.from_file(pkg, fn, filesystem)


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_build_format_database(filesystem):
    paths = [
        '/opt/ros/melodic/share/tf2_msgs',
        '/opt/ros/melodic/share/tf'
    ]
    db_package = PackageDatabase.from_paths(filesystem, paths)
    db_format = FormatDatabase.build(db_package)
    name_messages: Set[str] = set(db_format.messages)
    name_services: Set[str] = set(db_format.services)
    name_actions: Set[str] = set(db_format.actions)
    assert name_messages == {
        'tf/tfMessage',
        'tf/FrameGraphResponse',
        'tf2_msgs/FrameGraphResponse',
        'tf2_msgs/LookupTransformAction',
        'tf2_msgs/LookupTransformActionFeedback',
        'tf2_msgs/LookupTransformActionGoal',
        'tf2_msgs/LookupTransformActionResult',
        'tf2_msgs/LookupTransformFeedback',
        'tf2_msgs/LookupTransformGoal',
        'tf2_msgs/LookupTransformResult',
        'tf2_msgs/TF2Error',
        'tf2_msgs/TFMessage'
    }
    assert name_services == {
        'tf/FrameGraph',
        'tf2_msgs/FrameGraph'
    }
    assert name_actions == {'tf2_msgs/LookupTransform'}


@pytest.mark.parametrize('filesystem', ['fetch'], indirect=True)
def test_msg_toposort(filesystem):
    paths = [
        '/ros_ws/src/geometry2/tf2_msgs',
        '/ros_ws/src/geometry/tf',
        '/ros_ws/src/common_msgs/geometry_msgs',
        '/ros_ws/src/std_msgs'
    ]
    db_package = PackageDatabase.from_paths(filesystem, paths)
    db_format = FormatDatabase.build(db_package)

    msgs = db_format.messages.values()
    msgs = MsgFormat.toposort(msgs)


def test_msg_flatten():
    register: Dict[str, MsgFormat] = {}

    def mf(name: str, definition: str) -> MsgFormat:
        f = MsgFormat.from_string('example_pkg', name, definition)
        register[f.fullname] = f
        return f

    def names(fmt: MsgFormat) -> List[str]:
        buff = []
        for ctx, field in fmt.flatten(register):
            name = field.name
            if ctx:
                name = f"{'.'.join(ctx)}.{name}"
            buff.append(name)
        return buff

    f1 = mf("Item", """
uint32 foo
duration    lifespan
    """)
    f2 = mf("Container", """
Item        prize
Item[]      contents
time        stamp
    """)

    assert names(f1) == ['foo', 'lifespan']
    assert names(f2) == ['prize.foo', 'prize.lifespan', 'contents', 'stamp']



# -*- coding: utf-8 -*-
__all__ = ('ActionFormat',)

from typing import Optional, List, Dict, Any
import os

import attr
import dockerblade

from .msg import MsgFormat
from .. import exceptions


@attr.s(frozen=True, auto_attribs=True, slots=True)
class ActionFormat:
    """Provides an immutable definition of a
    `ROS Action <https://www.christimperley.co.uk/roswire/>`_.

    Attributes
    ----------
    package: str
        The name of the package to which this action belongs.
    name: str
        The name of this action.
    definition: str
        The plaintext definition of this action (e.g., the contents of a
        .action file).
    goal: MsgFormat
        The definition of the goal message for this action.
    feedback: Optional[MsgFormat]
        The definition of the optional feedback message for this action, if
        it has one.
    result: Optional[MsgFormat]
        The definition of the optional result message for this action, if
        it has one.
    """
    package: str
    name: str
    definition: str
    goal: MsgFormat
    feedback: Optional[MsgFormat]
    result: Optional[MsgFormat]

    @staticmethod
    def from_file(package: str,
                  filename: str,
                  files: dockerblade.FileSystem
                  ) -> 'ActionFormat':
        """Constructs an action format from a .action file for a given package.

        Parameters
        ----------
        package: str
            The name of the package that provides the file.
        filename: str
            The path to the .msg file.
        files: dockerblade.FileSystem
            An interface to the filesystem that hosts the .action file.

        Raises
        ------
        FileNotFoundError
            If the given file cannot be found.
        """
        assert filename.endswith('.action'), \
            'action format files must end in .action'
        name: str = os.path.basename(filename[:-7])
        contents: str = files.read(filename)
        return ActionFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, s: str) -> 'ActionFormat':
        """Constructs an action format from its definition (i.e., the contents
        of a .action file).

        Raises
        ------
        ParsingError
            If the description cannot be parsed.
        """
        goal: MsgFormat
        feed: Optional[MsgFormat]
        res: Optional[MsgFormat]

        name_goal = f"{name}Goal"
        name_feed = f"{name}Feedback"
        name_res = f"{name}Result"

        sections: List[str] = [ss.strip() for ss in s.split('---')]
        try:
            s_goal, s_res, s_feed = sections
        except ValueError:
            m = "failed to parse action description: expected three sections."
            raise exceptions.ParsingError(m)

        goal = MsgFormat.from_string(package, name_goal, s_goal)
        feed = MsgFormat.from_string(package, name_feed, s_feed)
        res = MsgFormat.from_string(package, name_res, s_res)
        return ActionFormat(package, name, s, goal, feed, res)

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None
                  ) -> 'ActionFormat':
        name: str = d['name']
        definition: str = d['definition']
        if package is None:
            assert d['package'] is not None
            package = d['package']

        res: Optional[MsgFormat] = None
        feed: Optional[MsgFormat] = None
        goal: MsgFormat = \
            MsgFormat.from_dict(d['goal'], package=package, name=f'{name}Goal')

        if 'result' in d:
            res = MsgFormat.from_dict(d['result'],
                                      package=package,
                                      name=f'{name}Result')
        if 'feedback' in d:
            feed = MsgFormat.from_dict(d['feedback'],
                                       package=package,
                                       name=f'{name}Feedback')

        return ActionFormat(package, name, definition, goal, feed, res)

    def to_dict(self) -> Dict[str, Any]:
        d = {'package': self.package,
             'name': self.name,
             'definition': self.definition,
             'goal': self.goal.to_dict()}
        if self.feedback:
            d['feedback'] = self.feedback.to_dict()
        if self.result:
            d['result'] = self.result.to_dict()
        return d

    @property
    def fullname(self) -> str:
        """The fully qualified name of this action."""
        return f"{self.package}/{self.name}"



# -*- coding: utf-8 -*-
from .base import Time, Duration
from .msg import Constant, ConstantValue, Field, MsgFormat, Message
from .srv import SrvFormat
from .action import ActionFormat
from .package import Package, PackageDatabase
from .format import FormatDatabase
from .type_db import TypeDatabase



# -*- coding: utf-8 -*-
__all__ = ('ROSLaunchManager',)

from typing import Collection, List, Mapping, Optional, Sequence, Tuple, Union
import os
import shlex
import xml.etree.ElementTree as ET

from loguru import logger
import attr
import dockerblade

from .config import LaunchConfig
from .controller import ROSLaunchController
from .reader import LaunchFileReader
from ... import exceptions as exc


@attr.s(eq=False)
class ROSLaunchManager:
    """Provides access to `roslaunch <wiki.ros.org/roslaunch/>`_ for an
    associated ROS system. This interface is used to locate, read, and write
    `launch XML files <http://wiki.ros.org/roslaunch/XML>`_,
    and to launch ROS nodes using those files.
    """
    _shell: dockerblade.shell.Shell = attr.ib(repr=False)
    _files: dockerblade.files.FileSystem = attr.ib(repr=False)

    def read(self,
             filename: str,
             *,
             package: Optional[str] = None,
             argv: Optional[Sequence[str]] = None
             ) -> LaunchConfig:
        """Produces a summary of the effects of a launch file.

        Parameters
        ----------
        filename: str
            The name of the launch file, or an absolute path to the launch
            file inside the container.
        package: str, optional
            The name of the package to which the launch file belongs.
        argv: Sequence[str], optional
            An optional sequence of command-line arguments that should be
            supplied to :code:`roslaunch`.

        Raises
        ------
        PackageNotFound
            If the given package could not be found.
        LaunchFileNotFound
            If the given launch file could not be found in the package.
        """
        filename = self.locate(filename, package=package)
        reader = LaunchFileReader(shell=self._shell, files=self._files)
        return reader.read(filename, argv)

    def write(self,
              config: LaunchConfig,
              *,
              filename: Optional[str] = None
              ) -> str:
        """Writes a given launch configuration to disk as an XML launch file.

        Parameters
        ----------
        config: LaunchConfig
            A launch configuration.
        filename: str, optional
            The name of the file to which the configuration should be written.
            If no filename is given, a temporary file will be created. It is
            the responsibility of the caller to ensure that the temporary file
            is appropriately destroyed.

        Returns
        -------
        str
            The absolute path to the generated XML launch file.
        """
        if not filename:
            filename = self._files.mktemp(suffix='.xml.launch')
        contents = ET.tostring(config.to_xml_tree().getroot())
        self._files.write(filename, contents)
        return filename  # type: ignore

    def locate(self, filename: str, *, package: Optional[str] = None) -> str:
        """Locates a given launch file.

        Parameters
        ----------
        filename: str
            The name of the launch file, or an absolute path to the launch
            file inside the container.
        package: str, optional
            Optionally specifies the name of the package to which the launch
            file belongs.

        Returns
        -------
        The absolute path to the launch file, if it exists.

        Raises
        ------
        PackageNotFound
            If the given package could not be found.
        LaunchFileNotFound
            If the given launch file could not be found in the package.
        """
        if not package:
            assert os.path.isabs(filename)
            return filename

        filename_original = filename
        logger.debug(f'determing location of launch file [{filename}]'
                     f' in package [{package}]')
        command = f'rospack find {shlex.quote(package)}'
        try:
            package_path = self._shell.check_output(command,
                                                    stderr=False,
                                                    text=True)
        except dockerblade.CalledProcessError as err:
            raise exc.PackageNotFound(package) from err
        filename = os.path.join(package_path, 'launch', filename_original)
        if not self._files.isfile(filename):
            raise exc.LaunchFileNotFound(path=filename)
        logger.debug('determined location of launch file'
                     f' [{filename_original}] in package [{package}]: '
                     f'{filename}')
        return filename

    def launch(self,
               filename: str,
               *,
               package: Optional[str] = None,
               args: Optional[Mapping[str, Union[int, str]]] = None,
               prefix: Optional[str] = None,
               launch_prefixes: Optional[Mapping[str, str]] = None,
               node_to_remappings: Optional[Mapping[str, Collection[Tuple[str, str]]]] = None  # noqa
               ) -> ROSLaunchController:
        """Provides an interface to the roslaunch command.

        Parameters
        ----------
        filename: str
            The name of the launch file, or an absolute path to the launch
            file inside the container.
        package: str, optional
            The name of the package to which the launch file belongs.
        args: Dict[str, Union[int, str]], optional
            Keyword arguments that should be supplied to roslaunch.
        prefix: str, optional
            An optional prefix to add before the roslaunch command.
        launch_prefixes: Mapping[str, str], optional
            An optional mapping from nodes, given by their names, to their
            individual launch prefix.
        node_to_remappings: Mapping[str, Collection[Tuple[str, str]]], optional
            A collection of name remappings for each node, represented as a
            mapping from node names to a collection of remappings for that
            node, where each remapping is a tuple of the
            form :code:`(to, from)`.

        Returns
        -------
        ROSLaunchController
            An interface for inspecting and managing the launch process.

        Raises
        ------
        PackageNotFound
            If the given package could not be found.
        LaunchFileNotFound
            If the given launch file could not be found in the package.
        """
        shell = self._shell
        if not args:
            args = {}
        if not launch_prefixes:
            launch_prefixes = {}
        filename = self.locate(filename, package=package)

        if node_to_remappings or launch_prefixes:
            launch_config = self.read(filename, package=package)
            logger.debug(f'instrumenting launch config: {launch_config}')

            if launch_prefixes:
                m = "individual launch prefixes are not yet implemented"
                raise NotImplementedError(m)

            if node_to_remappings:
                logger.debug('adding remappings to launch config: '
                             f'{node_to_remappings}')
                launch_config = \
                    launch_config.with_remappings(node_to_remappings)
                logger.debug('added remappings to launch config: '
                             f'{launch_config}')

            # write the instrumented launch config to a temporary file inside
            # the container and use that container with roslaunch
            package = None
            filename = self._files.mktemp(suffix='.launch')
            contents = launch_config.to_xml_string()
            logger.debug(f'instrumented launch file [{filename}]:\n{contents}')
            self._files.write(filename, contents)

        cmd = ['roslaunch', shlex.quote(filename)]
        launch_args: List[str] = [f'{arg}:={val}' for arg, val in args.items()]
        cmd += launch_args
        if prefix:
            cmd = [prefix] + cmd
        cmd_str = ' '.join(cmd)
        popen = shell.popen(cmd_str, stdout=True, stderr=True)

        return ROSLaunchController(filename=filename,
                                   popen=popen)

    __call__ = launch



# -*- coding: utf-8 -*-
"""
This module provides handling of XML substitution arguments, which are used
by XML launch and xacro files.

Reference
---------
https://github.com/ros/ros_comm/tree/kinetic-devel/tools/roslaunch/src/roslaunch/substitution_args.py
"""
__all__ = ('ArgumentResolver',)

from typing import Any, Dict, Match
import os
import re
import shlex
import subprocess

from loguru import logger
import attr
import dockerblade

from ...exceptions import SubstitutionError

R_ARG = re.compile(r'\$\(.+?\)')
R_FIND_ARG = re.compile(r'\$\(find .+?\)([^\s]+)')


@attr.s(auto_attribs=True)
class ArgumentResolver:
    shell: dockerblade.Shell
    files: dockerblade.FileSystem
    context: Dict[str, Any] = attr.ib(default=None)

    def _resolve_substitution_arg(self, s: str) -> str:
        """
        Raises
        ------
        EnvNotFoundError
            if a given environment variable is not found.
        """
        logger.debug(f"resolving substitution argument: {s}")
        s = s[2:-1]
        logger.debug(f"stripped delimiters: {s}")
        kind, *params = s.split(' ')
        logger.debug(f"argument kind: {kind}")

        # we deal with find in a later stage
        if kind == 'find':
            return f'$({s})'
        elif kind == 'env':
            var = params[0]
            return self._resolve_env(var)
        elif kind == 'optenv':
            var = params[0]
            default = ' '.join(params[1:])
            return self._resolve_optenv(var, default)
        elif kind == 'dirname':
            return self._resolve_dirname()
        elif kind == 'arg':
            arg_name = params[0]
            return self._resolve_arg(arg_name)
        elif kind == 'anon':
            return self._resolve_anon(params[0])
        return s

    def _resolve_dirname(self) -> str:
        try:
            dirname = os.path.dirname(self.context['filename'])
        except KeyError:
            m = 'filename is not provided by the launch context'
            raise SubstitutionError(m)
        return os.path.normpath(dirname)

    def _resolve_anon(self, name: str) -> str:
        raise NotImplementedError

    def _resolve_env(self, var: str) -> str:
        return self.shell.environ(var)

    def _resolve_optenv(self, var: str, default: str) -> str:
        try:
            return self.shell.environ(var)
        except dockerblade.exceptions.EnvNotFoundError:
            return default

    def _resolve_arg(self, arg_name: str) -> str:
        context = self.context
        if 'arg' not in context or arg_name not in context['arg']:
            m = f'arg not supplied to launch context [{arg_name}]'
            raise SubstitutionError(m)
        return context['arg'][arg_name]

    def _find_package_path(self, package: str) -> str:
        cmd = f'rospack find {shlex.quote(package)}'
        try:
            location = self.shell.check_output(cmd, text=True)
        except subprocess.CalledProcessError as err:
            raise SubstitutionError(f'failed to locate package: {package}') from err  # noqa
        return location.strip()

    def _find_executable(self, package: str, path: str) -> str:
        logger.debug(f'$(find-executable [package={package}] [path={path}]')
        path_original = path

        # look for executable in lib/ directory of workspaces
        catkin_find_command = ("catkin_find --first-only --libexec "
                               f"{shlex.quote(package)} {shlex.quote(path)}")
        try:
            path_in_ws = self.shell.check_output(catkin_find_command,
                                                 text=True)
            path_in_ws = path_in_ws.strip()
            return path_in_ws
        except dockerblade.CalledProcessError:
            pass

        # look for executable in source directory of package
        path_package = self._find_package_path(package)
        path_in_package = os.path.join(path_package, path)
        if not self.files.access(path_in_package, os.X_OK):
            m = ("$(find-executable pkg path) failed "
                 f"[package={package}; path={path_original}]")
            raise SubstitutionError(m)
        return path_in_package

    def _find_resource(self, package: str, path: str) -> str:
        catkin_find_command = ("catkin_find --first-only --share "
                               f"{shlex.quote(package)} {shlex.quote(path)}")
        try:
            path_in_ws = self.shell.check_output(catkin_find_command,
                                                 text=True)
            path_in_ws = path_in_ws.strip()
            return path_in_ws
        except dockerblade.CalledProcessError:
            pass

        path_package = self._find_package_path(package)
        path_in_package = os.path.join(path_package, path)
        if not self.files.isfile(path_in_package):
            m = ("$(find-resource pkg path) failed "
                 f"[package={package}; path={path}]")
            raise SubstitutionError(m)
        return path_in_package

    def _resolve_find(self, package: str, path: str = '') -> str:
        logger.debug(f'resolving find: {package}')
        path_original = path

        # if there is a path following the $(find ...) argument, then we first
        # attempt to treat the argument as part of an executable or resource
        # path
        if path:
            path = path.replace('\\', '/')
            if path.startswith('/'):
                path = path[1:]

            try:
                resolved_path = self._find_executable(package, path)
                logger.debug(f'resolved executable path: {resolved_path}')
                return resolved_path
            except SubstitutionError:
                pass
            try:
                resolved_path = self._find_resource(package, path)
                logger.debug(f'resolved resource path: {resolved_path}')
                return resolved_path
            except SubstitutionError:
                pass

        resolved_path = self._find_package_path(package) + path_original
        return resolved_path

    def _resolve_eval(self, attribute_string: str) -> str:
        logger.debug(f'resolving eval: {attribute_string}')
        assert attribute_string.startswith('$(eval ')
        assert attribute_string[-1] == ')'
        eval_string = attribute_string[7:-1]

        if '__' in attribute_string:
            m = ("$(eval ...): refusing to evaluate potentially dangerous "
                 "expression -- must not contain double underscores")
            raise SubstitutionError(m)

        _builtins = {x: __builtins__[x]  # type: ignore
                     for x in ('dict', 'float', 'int', 'list', 'map')}
        _locals = {
            'true': True,
            'True': True,
            'false': False,
            'False': False,
            '__builtins__': _builtins,
            'arg': self._resolve_arg,
            'anon': self._resolve_anon,
            'dirname': self._resolve_dirname,
            'env': self._resolve_env,
            'find': self._resolve_find,
            'optenv': self._resolve_optenv
        }

        result = str(eval(eval_string, {}, _locals))
        logger.debug(f'resolved eval [{attribute_string}]: {result}')
        return result

    def resolve(self, s: str) -> str:
        """Resolves a given argument string."""
        if s.startswith('$(eval ') and s[-1] == ')':
            return self._resolve_eval(s)
        s = R_ARG.sub(lambda m: self._resolve_substitution_arg(m.group(0)), s)

        def process_find_arg(match: Match[str]) -> str:
            # split tag and optional trailing path
            tag: str
            path: str
            tag, path = match.group(0).split(')', 1)
            tag += ')'
            args = tag[2:-1].split(' ')
            assert len(args) == 2
            tag_name, package = args
            assert tag_name == 'find'
            return self._resolve_find(package, path)

        # resolve find arguments
        s = R_FIND_ARG.sub(process_find_arg, s)
        return s



# -*- coding: utf-8 -*-
__all__ = ('Package', 'PackageDatabase')

from typing import Tuple, List, Dict, Any, Iterator, Collection, Mapping
import os

from loguru import logger
import attr
import dockerblade

from .msg import MsgFormat
from .srv import SrvFormat
from .action import ActionFormat
from ..util import tuple_from_iterable


@attr.s(frozen=True, auto_attribs=True, slots=True)
class Package:
    name: str
    path: str
    messages: Tuple[MsgFormat, ...] = attr.ib(converter=tuple_from_iterable)
    services: Tuple[SrvFormat, ...] = attr.ib(converter=tuple_from_iterable)
    actions: Tuple[ActionFormat, ...] = attr.ib(converter=tuple_from_iterable)

    @staticmethod
    def build(path: str, files: dockerblade.FileSystem) -> 'Package':
        """Constructs a description of a package at a given path."""
        name: str = os.path.basename(path)
        messages: List[MsgFormat] = []
        services: List[SrvFormat] = []
        actions: List[ActionFormat] = []

        if not files.isdir(path):
            raise FileNotFoundError(f"directory does not exist: {path}")

        dir_msg = os.path.join(path, 'msg')
        dir_srv = os.path.join(path, 'srv')
        dir_action = os.path.join(path, 'action')

        if files.isdir(dir_msg):
            messages = [MsgFormat.from_file(name, f, files)
                        for f in files.listdir(dir_msg, absolute=True)
                        if f.endswith('.msg')]
        if files.isdir(dir_srv):
            services = [SrvFormat.from_file(name, f, files)
                        for f in files.listdir(dir_srv, absolute=True)
                        if f.endswith('.srv')]
        if files.isdir(dir_action):
            actions = [ActionFormat.from_file(name, f, files)
                       for f in files.listdir(dir_action, absolute=True)
                       if f.endswith('.action')]

        return Package(name,  # type: ignore
                       path,
                       messages,
                       services,
                       actions)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Package':
        name: str = d['name']
        messages: List[MsgFormat] = [MsgFormat.from_dict(dd, package=name)
                                     for dd in d.get('messages', [])]
        services: List[SrvFormat] = [SrvFormat.from_dict(dd, package=name)
                                     for dd in d.get('services', [])]
        actions: List[ActionFormat] = [ActionFormat.from_dict(dd, package=name)
                                       for dd in d.get('actions', [])]
        return Package(d['name'],  # type: ignore
                       d['path'],
                       messages,
                       services,
                       actions)

    def to_dict(self) -> Dict[str, Any]:
        d = {'name': self.name,
             'path': self.path,
             'messages': [m.to_dict() for m in self.messages],
             'services': [s.to_dict() for s in self.services],
             'actions': [a.to_dict() for a in self.actions]}
        return d


class PackageDatabase(Mapping[str, Package]):
    """
    An immutable database of packages, represented as :class:`Package`
    instances, indexed by their names, given as :class:`str`.

    Note
    ----
        Implements most :class:`dict` operations via
        :class:`abc.collections.Mapping`,
        including :code:`db['name']`, :code:`len(db)`), :code:`db.keys()`,
        :code:`db.values()`, and :code:`iter(db)`.
        As instances of this class are immutable, no destructive
        :class:`dict` operations are provided (e.g., :code:`del db['foo'])`
        and `db['foo'] = bar`).
    """
    @staticmethod
    def paths(shell: dockerblade.Shell,
              files: dockerblade.FileSystem
              ) -> List[str]:
        """Parses :code:`ROS_PACKAGE_PATH` for a given shell."""
        path_str = shell.environ('ROS_PACKAGE_PATH')
        package_paths: List[str] = path_str.strip().split(':')
        paths: List[str] = []
        for path in package_paths:
            try:
                all_packages = files.find(path, 'package.xml')
            except dockerblade.exceptions.DockerBladeException:
                logger.warning('unable to find directory in ROS_PACKAGE_PATH:'
                               f' {path}')
                continue
            package_dirs = [os.path.dirname(p) for p in all_packages]
            paths.extend(package_dirs)
        return paths

    @staticmethod
    def from_paths(files: dockerblade.FileSystem,
                   paths: List[str],
                   ignore_bad_paths: bool = True
                   ) -> 'PackageDatabase':
        """
        Constructs a package database from a list of the paths of the packages
        belonging to the database.

        Parameters
        ----------
        files: dockerblade.FileSystem
            access to the filesystem.
        paths: List[str]
            a list of the absolute paths of the packages.
        ignore_bad_paths: bool
            If :code:`True`, non-existent paths will be ignored.
            If :code:`False`, a :exc:`FileNotFoundError` will be raised.

        Raises
        ------
        FileNotFoundError
            if no package is found at a given path.
        """
        packages: List[Package] = []
        for p in paths:
            try:
                package = Package.build(p, files)
            except FileNotFoundError:
                logger.exception(f"unable to build package: {p}")
                if not ignore_bad_paths:
                    raise
            else:
                packages.append(package)
        return PackageDatabase(packages)

    def __init__(self, packages: Collection[Package]) -> None:
        self.__contents = {p.name: p for p in packages}

    def __len__(self) -> int:
        """Returns the number of packages within this database."""
        return len(self.__contents)

    def __getitem__(self, name: str) -> Package:
        """Fetches the description for a given package.

        Raises
        ------
        KeyError
            if no package exists with the given name.
        """
        return self.__contents[name]

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of the packages contained within
        this database.
        """
        yield from self.__contents

    @staticmethod
    def from_dict(d: List[Dict[str, Any]]) -> 'PackageDatabase':
        return PackageDatabase([Package.from_dict(dd) for dd in d])

    def to_dict(self) -> List[Dict[str, Any]]:
        return [p.to_dict() for p in self.values()]



# -*- coding: utf-8 -*-
__all__ = ('ParameterServer',)

from typing import Any, Iterator, Mapping, Sequence
import xmlrpc.client


from .. import exceptions


class ParameterServer(Mapping[str, Any]):
    """
    See: http://wiki.ros.org/ROS/Parameter%20Server%20API
    """
    def __init__(self, connection: xmlrpc.client.ServerProxy) -> None:
        """
        Constructs a new parameter server proxy using an XML-RPC server proxy
        for a given ROS master.
        """
        self.__caller_id = '/.roswire'
        self.__connection = connection

    def __len__(self) -> int:
        """
        Returns a count of the number of registered parameters.
        """
        return len(list(self))

    def __contains__(self, key: object) -> bool:
        """
        Determines whether the parameter server contains a given parameter
        or tree with a given key.
        """
        assert isinstance(key, str)
        conn = self.__connection
        code, msg, result = \
            conn.hasParam(self.__caller_id, key)  # type: ignore
        if code != 1:
            raise exceptions.ROSWireException("bad API call!")
        assert isinstance(result, bool)
        return result

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of the parameters stored on the
        server.
        """
        code: int
        msg: str
        result: Sequence[str]
        code, msg, result = \
            self.__connection.getParamNames(self.__caller_id)  # type: ignore
        if code != 1:
            raise exceptions.ROSWireException("bad API call!")
        yield from result

    def __getitem__(self, key: str) -> Any:
        """
        Fetches the value of a given parameter from the server.
        If the provided key is a namespace, then the contents of that
        namespace will be returned as a dictionary.

        Parameters:
            key: the name of the parameter (or namespace).

        Returns:
            The value of the parameter or the contents of the given namespace.

        Raises:
            ParameterNotFoundError: if no parameter with the given key is found
                on the parameter server.
        """
        code, msg, result = \
            self.__connection.getParam(self.__caller_id, key)  # type: ignore
        if code == -1:
            raise exceptions.ParameterNotFoundError(key)
        if code != 1:
            raise exceptions.ROSWireException("bad API call")
        return result

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Sets the value of a parameter on the server. If the value is a
        dictionary, it will be treated as a parameter tree.
        """
        code, msg, result = \
            self.__connection.setParam(self.__caller_id, key, value)  # type: ignore  # noqa
        if code != 1:
            raise exceptions.ROSWireException("bad API call!")

    def __delitem__(self, key: str) -> None:
        """
        Deletes a given parameter (or parameter tree) from the server.

        Parameters:
            key: the key for the parameter or parameter tree.

        Raises:
            KeyError: if no parameter or parameter tree is found with the
                given key on the server.
        """
        code, msg, result = \
            self.__connection.deleteParam(self.__caller_id, key)  # type: ignore  # noqa
        if code == -1:
            raise exceptions.ParameterNotFoundError(key)
        if code != 1:
            raise exceptions.ROSWireException("bad API call!")



# -*- coding: utf-8 -*-
__all__ = ('Constant', 'ConstantValue', 'Field', 'MsgFormat', 'Message')

from typing import (Optional, Any, Union, Tuple, List, Dict, ClassVar,
                    Collection, Set, Iterator, Mapping, BinaryIO)
from io import BytesIO
import hashlib
import re
import os

from loguru import logger
from toposort import toposort_flatten as toposort
import attr
import dockerblade

from .base import is_builtin, Time, Duration
from .decode import is_simple
from .. import exceptions as exc

R_TYPE = r"[a-zA-Z0-9_/]+(?:\[\d*\])?"
R_NAME = r"[a-zA-Z0-9_/]+"
R_VAL = r".+"
R_COMMENT = r"(#.*)?"
R_FIELD = re.compile(f"^\s*({R_TYPE})\s+({R_NAME})\s*{R_COMMENT}$")
R_STRING_CONSTANT = re.compile("^\s*string\s+(\w+)\s*=\s*(.+)\s*$")
R_OTHER_CONSTANT = re.compile("^\s*(\w+)\s+(\w+)\s*=\s*([^\s]+).*$")
R_BLANK = re.compile(f"^\s*{R_COMMENT}$")

ConstantValue = Union[str, int, float]


@attr.s(frozen=True, slots=True, str=False, auto_attribs=True)
class Constant:
    """Provides an immutable definition of a constant for a message format.

    Attributes
    ----------
    typ: str
        The name of the type used by this constant.
    name: str
        The name of this constant.
    value: Union[str, int, float]
        The value of this constant.
    """
    typ: str
    name: str
    value: Union[str, int, float]

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Constant':
        return Constant(d['type'], d['name'], d['value'])

    def to_dict(self) -> Dict[str, Any]:
        return {'type': self.typ,
                'name': self.name,
                'value': self.value}

    def __str__(self) -> str:
        return f"{self.typ} {self.name}={str(self.value)}"


@attr.s(frozen=True, str=False, slots=True, auto_attribs=True)
class Field:
    """Provides an immutable description of a message field.

    Attributes
    ----------
    typ: str
        The name of the type used this field.
    name: str
        The name of this field.
    """
    typ: str
    name: str

    @property
    def is_array(self) -> bool:
        return '[' in self.typ

    @property
    def is_simple(self) -> bool:
        return not self.is_array and is_simple(self.typ)

    @property
    def length(self) -> Optional[int]:
        if not self.is_array:
            return None
        sz = self.typ.partition('[')[2].partition(']')[0]
        if sz == '':
            return None
        return int(sz)

    @property
    def base_type(self) -> str:
        return self.typ.partition('[')[0] if self.is_array else self.typ

    @property
    def base_typ(self) -> str:
        return self.base_type

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Field':
        return Field(d['type'], d['name'])

    def to_dict(self) -> Dict[str, str]:
        return {'type': self.typ,
                'name': self.name}

    def without_package_name(self) -> 'Field':
        typ = self.typ.partition('/')[2] if '/' in self.typ else self.typ
        return Field(typ, self.name)

    def __str__(self) -> str:
        return f"{self.typ} {self.name}"


@attr.s(frozen=True, slots=True, auto_attribs=True)
class MsgFormat:
    """Provides an immutable definition of a given ROS message format.

    Attributes
    ----------
    package: str
        The name of the package that defines this message format.
    name: str
        The unqualified name of the message format.
    definition: str
        The plaintext contents of the associated .msg file.
    fields: Sequence[Field]
        The fields that belong to this message format.
    constants: Sequence[Constant]
        The named constants that belong to this message format.

    References
    ----------
    * http://wiki.ros.org/msg
    """
    package: str
    name: str
    definition: str
    fields: Tuple[Field, ...] = attr.ib(converter=tuple)
    constants: Tuple[Constant, ...] = attr.ib(converter=tuple)

    @staticmethod
    def toposort(fmts: Collection['MsgFormat']) -> List['MsgFormat']:
        fn_to_fmt: Dict[str, MsgFormat] = {fmt.fullname: fmt for fmt in fmts}
        fn_to_deps: Dict[str, Set[str]] = \
            {filename: {f.base_typ for f in fmt.fields
                        if not is_builtin(f.base_typ)}
             for filename, fmt in fn_to_fmt.items()}
        toposorted = list(toposort(fn_to_deps))
        missing_packages: Set[str] = set(toposorted) - set(fn_to_fmt)
        if missing_packages:
            logger.error("Messages are relied upon but are not present in "
                         f"the format database: {', '.join(missing_packages)}")
            missing_package_name = next(iter(missing_packages))
            raise exc.PackageNotFound(missing_package_name)
        return [fn_to_fmt[filename] for filename in toposorted]

    @staticmethod
    def from_file(package: str,
                  filename: str,
                  files: dockerblade.FileSystem
                  ) -> 'MsgFormat':
        """Constructs a message format from a .msg file for a given package.

        Parameters
        ----------
        package: str
            The name of the package that provides the file.
        filename: str
            The absolute path to the .msg file inside the given filesystem.
        files: dockerblade.FileSystem
            An interface to the filesystem that hosts the .msg file.

        Raises
        ------
        FileNotFoundError
            If the given file cannot be found.
        """
        assert filename.endswith('.msg'), \
            'message format files must end in .msg'
        name: str = os.path.basename(filename[:-4])
        contents: str = files.read(filename)
        return MsgFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, text: str) -> 'MsgFormat':
        """Constructs a message format from its description.

        Parameters
        ----------
        package: str
            The name of the package that provides the file.
        filename: str
            The absolute path to the .msg file inside the given filesystem.
        text: str
            The message definition itself (e.g., the contents of a .msg file).

        Raises
        ------
        ParsingError
            If the description cannot be parsed.
        """
        typ: str
        name_const: str
        fields: List[Field] = []
        constants: List[Constant] = []

        for line in text.split('\n'):
            m_blank = R_BLANK.match(line)
            m_string_constant = R_STRING_CONSTANT.match(line)
            m_other_constant = R_OTHER_CONSTANT.match(line)
            m_field = R_FIELD.match(line)

            if m_blank:
                continue
            elif m_string_constant:
                name_const, val = m_string_constant.group(1, 2)
                constant = Constant('string', name_const, val)
                constants.append(constant)
            elif m_other_constant:
                typ, name_const, val_str = m_other_constant.group(1, 2, 3)
                val = val_str  # FIXME convert value
                constant = Constant(typ, name_const, val)
                constants.append(constant)
            elif m_field:
                typ, name_field = m_field.group(1, 2)

                # resolve the type of the field
                typ_resolved = typ
                base_typ = typ.partition('[')[0]
                if typ == 'Header':
                    typ_resolved = 'std_msgs/Header'
                elif '/' not in typ and not is_builtin(base_typ):
                    typ_resolved = f'{package}/{typ}'

                if typ != typ_resolved:
                    logger.debug(f"resolved type [{typ}]: {typ_resolved}")
                    typ = typ_resolved

                field: Field = Field(typ, name_field)
                fields.append(field)
            else:
                raise exc.ParsingError(f"failed to parse line: {line}")

        return MsgFormat(package, name, text, fields, constants)  # type: ignore  # noqa

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None,
                  name: Optional[str] = None
                  ) -> 'MsgFormat':
        if not package:
            package = d['package']
        if not name:
            name = d['name']
        definition = d['definition']
        fields = [Field.from_dict(dd) for dd in d.get('fields', [])]
        constants = [Constant.from_dict(dd) for dd in d.get('constants', [])]
        return MsgFormat(package, name, definition, fields, constants)  # type: ignore  # noqa

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {'package': self.package,
                             'name': self.name,
                             'definition': self.definition}
        if self.fields:
            d['fields'] = [f.to_dict() for f in self.fields]
        if self.constants:
            d['constants'] = [c.to_dict() for c in self.constants]
        return d

    @property
    def fullname(self) -> str:
        """The fully qualified name of this message format."""
        return f"{self.package}/{self.name}"

    def flatten(self,
                name_to_format: Mapping[str, 'MsgFormat'],
                ctx: Tuple[str, ...] = tuple()
                ) -> Iterator[Tuple[Tuple[str, ...], Field]]:
        for field in self.fields:
            if field.is_array or is_builtin(field.typ):
                yield (ctx, field)
            else:
                fmt = name_to_format[field.typ]
                yield from fmt.flatten(name_to_format, ctx + (field.name,))

    def md5text(self, name_to_msg: Mapping[str, 'MsgFormat']) -> str:
        """Computes the MD5 text for this format."""
        lines: List[str] = []
        lines += [str(c) for c in self.constants]
        for f in self.fields:
            if is_builtin(f.base_type):
                lines += [str(f.without_package_name())]
            else:
                f_md5 = name_to_msg[f.base_type].md5sum(name_to_msg)
                lines += [f'{f_md5} {f.name}']
        return '\n'.join(lines)

    def md5sum(self, name_to_msg: Mapping[str, 'MsgFormat']) -> str:
        """Computes the MD5 sum for this format."""
        logger.debug(f"generating md5sum: {self.fullname}")
        txt = self.md5text(name_to_msg)
        logger.debug(f"generated md5 text [{self.fullname}]:\n{txt}")
        md5sum = hashlib.md5(txt.encode('utf-8')).hexdigest()
        logger.debug(f"generated md5sum [{self.fullname}]: {md5sum}")
        return md5sum


class Message:
    """Each ROS message type has its own class that is dynamically generated
    by ROSWire at runtime. This is the base class that is used by all of those
    messages.

    Attributes
    ----------------
    format: MsgFormat
        The format used by this message.
    """
    format: ClassVar[MsgFormat]

    @staticmethod
    def _to_dict_value(val: Any) -> Any:
        typ = type(val)

        if typ in (Time, Duration) or issubclass(typ, Message):
            return val.to_dict()

        if typ in (list, tuple):
            if not val:
                return []
            typ_item = type(val[0])
            if typ_item == Time or issubclass(typ_item, Message):
                return [vv.to_dict() for vv in val]
            # includes (str, int, float)
            return list(val)

        # includes (str, int, float)
        return val

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {}
        for field in self.format.fields:
            name: str = field.name
            val = getattr(self, field.name)
            d[name] = self._to_dict_value(val)
        return d

    @classmethod
    def md5sum(cls) -> str:
        """Returns the md5sum for this message type."""
        raise NotImplementedError

    @classmethod
    def read(cls, b: BinaryIO) -> 'Message':
        """Reads a binary encoding of this message from a given stream."""
        raise NotImplementedError

    @classmethod
    def decode(cls, b: bytes) -> 'Message':
        """Decodes a binary encoding of this message."""
        return cls.read(BytesIO(b))

    def write(self, b: BinaryIO) -> None:
        """Writes a binary encoding of this message to a given stream."""
        raise NotImplementedError

    def encode(self) -> bytes:
        """Returns a binary encoding of this message."""
        b = BytesIO()
        self.write(b)
        return b.getvalue()



# -*- coding: utf-8 -*-
import pytest

from roswire.proxy import SystemState


@pytest.mark.parametrize('sut', ['fetch'], indirect=True)
def test_state(sut):
    with sut.roscore() as ros:
        expected_pubs = {
            '/rosout_agg': ['/rosout']
        }
        expected_subs = {
            '/rosout': ['/rosout']
        }
        expected_services = {
            '/rosout/set_logger_level': ['/rosout'],
            '/rosout/get_loggers': ['/rosout']
        }
        expected = SystemState(publishers=expected_pubs,
                               subscribers=expected_subs,
                               services=expected_services)
        assert ros.state == expected



# -*- coding: utf-8 -*-
__all__ = ('System',)

from typing import Iterator
from uuid import UUID
import contextlib

import attr
import dockerblade

from .description import SystemDescription
from .definitions import TypeDatabase
from .proxy import ROSCore, Container, CatkinInterface, CatkinTools, CatkinMake


@attr.s(frozen=True)
class System:
    """Provides access to a ROS application hosted by a Docker container.

    Attributes
    ----------
    uuid: UUID
        A unique identifier for the underlying container.
    description: SystemDescription
        A static description of the associated ROS application.
    messages: TypeDatabase
        A database of message types for the associated ROS application.
    shell: dockerblade.shell.Shell
        Provides access to a bash shell for this container.
    files: dockerblade.files.FileSystem
        Provides access to the filesystem for this container.
    ws_host: str
        The absolute path to the shared directory for this container's
        workspace on the host machine.
    container: Container
        Provides access to the underlying Docker container.
    """
    container: Container = attr.ib()
    description: SystemDescription = attr.ib()

    @property
    def uuid(self) -> UUID:
        return self.container.uuid

    @property
    def ws_host(self) -> str:
        return self.container.ws_host

    @property
    def ip_address(self) -> str:
        return self.container.ip_address

    @property
    def shell(self) -> dockerblade.shell.Shell:
        return self.container.shell

    @property
    def messages(self) -> TypeDatabase:
        return self.description.types

    def catkin(self, directory: str) -> CatkinInterface:
        """Returns an interface to a catkin workspace.

        Parameters
        ----------
        directory: str
            The absolute path to the catkin workspace inside the container.

        Returns
        -------
        CatkinInterface
            An interface to the given workspace.
        """
        # TODO decide whether to use catkin_tools or catkin_make
        return self.catkin_tools(directory)

    def catkin_tools(self, directory: str) -> CatkinTools:
        """Returns an interface to a catkin tools workspace.

        Parameters
        ----------
        directory: str
            The absolute path to the catkin workspace inside the container.

        Returns
        -------
        CatkinTools
            An interface to the given workspace.
        """
        return CatkinTools(shell=self.shell, directory=directory)

    def catkin_make(self, directory: str) -> CatkinMake:
        """Returns an interface to a catkin_make workspace.

        Parameters
        ----------
        directory: str
            The absolute path to the catkin workspace inside the container.

        Returns
        -------
        CatkinMake
            An interface to the given workspace.
        """
        return CatkinMake(shell=self.shell, directory=directory)

    @property
    def files(self) -> dockerblade.files.FileSystem:
        return self.container.files

    @contextlib.contextmanager
    def roscore(self, port: int = 11311) -> Iterator[ROSCore]:
        """
        Launches a context-managed roscore inside the container.
        Upon exiting the context, the ROS master (and its associated resources)
        will be destroyed.

        Parameters
        ----------
        port: int, optional
            The port inside the container on which roscore should run.

        Yields
        ------
        ROSCore
            An interface to the launched ROS Master.
        """
        assert port > 1023
        command = f"roscore -p {port}"
        process = self.shell.popen(command)
        try:
            yield ROSCore(description=self.description,
                          shell=self.shell,
                          files=self.files,
                          ws_host=self.ws_host,
                          ip_address=self.ip_address,
                          port=port)
        finally:
            process.terminate()
            process.wait(2.0)
            process.kill()



# -*- coding: utf-8 -*-
__all__ = ('LaunchConfig',)

from typing import (AbstractSet, Any, Collection, Dict, Mapping, Optional,
                    Set, Sequence, Tuple)
import xml.dom.minidom as minidom
import xml.etree.ElementTree as ET

from loguru import logger
import attr

from .node import NodeConfig
from .env import Env
from .parameter import Parameter
from ....exceptions import FailedToParseLaunchFile
from ....name import (canonical_name, name_is_global, namespace_join,
                      namespaces_of)


@attr.s(frozen=True, slots=True)
class LaunchConfig:
    nodes: AbstractSet[NodeConfig] = attr.ib(default=frozenset(),
                                             converter=frozenset)
    executables: Sequence[str] = attr.ib(default=tuple())
    roslaunch_files: Sequence[str] = attr.ib(default=tuple())
    params: Mapping[str, Any] = attr.ib(factory=dict)
    clear_params: Sequence[str] = attr.ib(default=tuple())
    errors: Sequence[str] = attr.ib(default=tuple())
    envs: Mapping[str, Env] = attr.ib(default=dict())

    def with_clear_param(self, ns: str) -> 'LaunchConfig':
        """Specifies a parameter that should be cleared before new parameters
        are set."""
        ns = canonical_name(ns)
        if ns in self.clear_params:
            return self
        clear_params = tuple(self.clear_params) + (ns,)
        return attr.evolve(self, clear_params=clear_params)

    def with_env(self,
                 name: str,
                 value: str
                 ) -> 'LaunchConfig':
        """Adds an environment variable to this configuration."""
        envs: Dict[str, Env] = dict(self.envs)
        envs[name] = Env(name=name, value=value)
        return attr.evolve(self, envs=envs)

    def with_rosparam(self,
                      name: str,
                      value: Any
                      ) -> 'LaunchConfig':
        # complex parameters
        if isinstance(value, dict):
            config: LaunchConfig = self
            for child_name, child_value in value.items():
                child_name = namespace_join(name, child_name)
                config = config.with_rosparam(child_name, child_value)
            return config

        # primitive parameters
        # FIXME better handling of types
        return self.with_param(name, typ='auto', value=value)

    def with_param(self,
                   name: str,
                   typ: str,
                   value: Any,
                   command: Optional[str] = None
                   ) -> 'LaunchConfig':
        """Adds a parameter to this configuration."""
        param = Parameter(name=name, typ=typ, value=value, command=command)
        params: Dict[str, Any] = dict(self.params)
        errors = self.errors

        if not name_is_global(name):
            m = f"expected parameter name to be global: {name}"
            raise FailedToParseLaunchFile(m)

        for parent_name in (n for n in namespaces_of(name) if n in params):
            err = f"parameter [{name}] conflicts with parent [{parent_name}]"
            errors = tuple(errors) + (err,)

        params[name] = param
        return attr.evolve(self, params=params, errors=errors)

    def with_remappings(self,
                        node_to_remappings: Mapping[str, Collection[Tuple[str, str]]]  # noqa
                        ) -> 'LaunchConfig':
        nodes: Set[NodeConfig] = set()
        for node in self.nodes:
            if node.name in node_to_remappings:
                node = node.with_remappings(node_to_remappings[node.name])
            nodes.add(node)
        return attr.evolve(self, nodes=frozenset(nodes))

    def with_executable(self, executable: str) -> 'LaunchConfig':
        """Specify an executable that should be run at launch."""
        executables = tuple(self.executables) + (executable,)
        return attr.evolve(self, executables=executables)

    def with_roslaunch_file(self, filename: str) -> 'LaunchConfig':
        roslaunch_files = tuple(self.roslaunch_files) + (filename,)
        return attr.evolve(self, roslaunch_files=roslaunch_files)

    def with_node(self, node: NodeConfig) -> 'LaunchConfig':
        logger.debug(f"adding node to config: {node}")
        used_names = {n.full_name for n in self.nodes}
        if node.full_name in used_names:
            m = 'multiple definitions of node [{}] in launch configuration'
            m = m.format(node.full_name)
            raise FailedToParseLaunchFile(m)
        nodes = self.nodes | frozenset({node})
        return attr.evolve(self, nodes=nodes)

    def to_xml_tree(self) -> ET.ElementTree:
        root = ET.Element('launch')
        for env in self.envs.values():
            root.append(env.to_xml_element())
        for param in self.params.values():
            root.append(param.to_xml_element())
        for node in self.nodes:
            root.append(node.to_xml_element())
        return ET.ElementTree(root)

    def to_xml_string(self) -> str:
        """Transforms this launch configuration to a launch XML file."""
        xml = self.to_xml_tree()
        contents = ET.tostring(xml.getroot())
        return minidom.parseString(contents).toprettyxml(indent='  ')

    def to_xml_file(self, filename: str) -> None:
        """Saves this launch configuration to an equivalent launch XML file."""
        contents = self.to_xml_string()
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(contents)



# -*- coding: utf-8 -*-
__all__ = ('SrvFormat',)

from typing import Optional, List, Dict, Any
import os

import attr
import dockerblade

from .msg import MsgFormat


@attr.s(frozen=True, auto_attribs=True, slots=True)
class SrvFormat:
    """Provides an immutable definition of a given
    `ROS service format <http://wiki.ros.org/srv>`_.

    Attributes
    ----------
    package: str
        The name of the package that defines this service format.
    name: str
        The unqualified name of the service format.
    definition: str
        The plaintext contents of the associated .srv file.
    request: Optional[MsgFormat]
        The definition of the optional request message for this service, if
        it has one.
    response: Optional[MsgFormat]
        The definition of the optional response message for this service, if
        it has one.
    """
    package: str
    name: str
    definition: str
    request: Optional[MsgFormat]
    response: Optional[MsgFormat]

    @staticmethod
    def from_file(package: str,
                  filename: str,
                  files: dockerblade.FileSystem
                  ) -> 'SrvFormat':
        """Constructs a service format from a .srv file for a given package.

        Parameters
        ----------
        package: str
            The name of the package that provides the file.
        filename: str
            The absolute path to the .srv file.
        files: dockerblade.FileSystem
            An interface to the filesystem that hosts the .srv file.

        Raises
        ------
        FileNotFoundError
            If the given file cannot be found.
        """
        assert filename.endswith('.srv'), \
            'service format files must end in .srv'
        name: str = os.path.basename(filename[:-4])
        contents: str = files.read(filename)
        return SrvFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, s: str) -> 'SrvFormat':
        """Constructs a service format from its definition.

        Raises
        ------
        ParsingError
            If the description cannot be parsed.
        """
        req: Optional[MsgFormat] = None
        res: Optional[MsgFormat] = None
        name_req = f"{name}Request"
        name_res = f"{name}Response"

        sections: List[str] = [ss.strip() for ss in s.split('---')]
        assert len(sections) < 3
        s_req = sections[0]
        s_res = sections[1] if len(sections) > 1 else ''

        if s_req:
            req = MsgFormat.from_string(package, name_req, s_req)
        if s_res:
            res = MsgFormat.from_string(package, name_res, s_res)

        return SrvFormat(package, name, s, req, res)  # type: ignore

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None
                  ) -> 'SrvFormat':
        req: Optional[MsgFormat] = None
        res: Optional[MsgFormat] = None
        name: str = d['name']
        definition: str = d['definition']
        if package is None:
            assert d['package'] is not None
            package = d['package']

        if 'request' in d:
            req = MsgFormat.from_dict(d['request'],
                                      package=package,
                                      name=f'{name}Request')
        if 'response' in d:
            res = MsgFormat.from_dict(d['response'],
                                      package=package,
                                      name=f'{name}Response')

        return SrvFormat(package, name, definition, req, res)

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {'package': self.package,
                             'name': self.name,
                             'definition': self.definition}
        if self.request:
            d['request'] = self.request.to_dict()
        if self.response:
            d['response'] = self.response.to_dict()
        return d

    @property
    def fullname(self) -> str:
        """The fully qualified name of this service format."""
        return f"{self.package}/{self.name}"
